theory Xmlt2
  imports
    "HOL-Library.Extended_Nat"
    "XML.Xmlt"
    Show.Show_Instances
    "Sqrt_Babylonian.Log_Impl"
    "HOL-Library.Code_Char"
begin

(*TODO: Not used anymore, but should be usefull. move ?
lemma sum_mono [partial_function_mono]:
  assumes p1: "\<And>x. mono_option (p1 x)"
      and p2: "\<And>x. mono_option (p2 x)"
  shows "mono_option (\<lambda>g. case x of Inl x \<Rightarrow> p1 x g | Inr x \<Rightarrow> p2 x g)"
  using assms by (cases x, auto intro!:partial_function_mono)

lemma prod_mono [partial_function_mono]:
  assumes "\<And>a b. mono_option (p a b)"
  shows "mono_option (\<lambda>g. case x of (a,b) \<Rightarrow> p a b g)"
  using assms by (cases x, auto intro!:partial_function_mono)

lemma option_mono [partial_function_mono]:
  assumes p1: "\<And>x. mono_option (p1 x)"
      and p2: "mono_option p2"
  shows "mono_option (\<lambda>g. case x of Some x \<Rightarrow> p1 x g | None \<Rightarrow> p2 g)"
  using assms by (cases x, auto intro!:partial_function_mono)
*)

datatype 'a xml_error = TagMismatch "tag list" | Fatal 'a
text {*
  @{term "TagMismatch tags"} represents tag mismatch, expecting one of @{term tags} but
  something else is encountered.
*}

lemma xml_error_mono [partial_function_mono]:
  assumes p1: "\<And>tags. mono_option (p1 tags)"
      and p2: "\<And>x. mono_option (p2 x)"
      and f: "mono_option f"
  shows "mono_option (\<lambda>g. case s of TagMismatch tags \<Rightarrow> p1 tags g | Fatal x \<Rightarrow> p2 x g)"
  using assms by (cases s, auto intro!:partial_function_mono)

text {* A state is a tuple of
  the XML or list of XMLs to be parsed,
  the attributes,
  a flag indicating if mismatch is allowed,
  a list of tags that have been mismatched,
  the current position.
 *}
type_synonym 'a xmlt2 = "xml \<times> (string \<times> string) list \<times> bool \<times> tag list \<times> tag list \<Rightarrow> string xml_error + 'a"
type_synonym 'a xmlst = "xml list \<times> (string \<times> string) list \<times> bool \<times> tag list \<times> tag list \<Rightarrow> string xml_error + 'a"

lemma xml_state_cases:
  assumes "\<And>p nam atts xmls. x = (XML nam atts xmls, p) \<Longrightarrow> thesis"
      and "\<And>p txt. x = (XML_text txt, p) \<Longrightarrow> thesis"
  shows thesis
  using assms by (cases x; cases "fst x", auto)

lemma xmls_state_cases:
  assumes "\<And>p. x = ([],p) \<Longrightarrow> thesis"
      and "\<And>xml xmls p. x = (xml # xmls, p) \<Longrightarrow> thesis"
  shows thesis
  using assms by (cases x; cases "fst x", auto)

lemma xmls_state_induct:
  fixes x :: "xml list \<times> _"
  assumes "\<And>a b c d. P ([],a,b,c,d)"
      and "\<And>xml xmls a b c d. (\<And>a b c d. P (xmls,a,b,c,d)) \<Longrightarrow> P (xml # xmls, a, b, c, d)"
  shows "P x"
proof (induct x)
  case (fields xmls a b c d)
  with assms show ?case by (induct xmls arbitrary:a b c d, auto)
qed

definition xml_error
where "xml_error str x \<equiv> case x of (xmls,_,_,_,pos) \<Rightarrow>
  let next = case xmls of
      XML tag _ _ # _ \<Rightarrow> ''<'' @ tag @ ''>''
    | XML_text str # _ \<Rightarrow> ''text element \"'' @ str @ ''\"''
    | [] \<Rightarrow> ''tag close''
  in
  Inl (Fatal (''parse error on '' @ next @ '' at '' @ show pos @ '':\<newline>'' @ str))"

definition xml_return :: "'a \<Rightarrow> 'a xmlst"
where "xml_return v x \<equiv> case x
  of ([],_) \<Rightarrow> Inr v
  | _ \<Rightarrow> xml_error (''expecting tag close'') x"

definition "mismatch tag x \<equiv> case x of
  (xmls,atts,flag,cands,_) \<Rightarrow>
    if flag then Inl (TagMismatch (tag#cands))
    else xml_error (''expecting '' @ show (tag#cands)) x"

abbreviation xml_any :: "xml xmlt2"
where
  "xml_any x \<equiv> Inr (fst x)"

text {* Conditional parsing depending on tag match. *}

definition xml_or (infixr "XMLor" 51)
where
 "xml_or p1 p2 x \<equiv> case x of (x1,atts,flag,cands,rest) \<Rightarrow> (
  case p1 (x1,atts,True,cands,rest)
  of Inl err1 \<Rightarrow> (
    case err1
    of TagMismatch cands1 \<Rightarrow> p2 (x1,atts,flag,cands1,rest)
    | err1 \<Rightarrow> Inl err1
  )
  | ret1 \<Rightarrow> ret1
)"

definition xml_do :: "tag \<Rightarrow> 'a xmlst \<Rightarrow> 'a xmlt2" where
 "xml_do tag p x \<equiv>
  case x of (XML nam atts xmls, _, flag, cands, pos) \<Rightarrow>
    if nam = tag then p (xmls,atts,False,[],tag#pos) (* inner tag mismatch is not allowed *)
    else mismatch tag ([fst x], snd x)
   | _ \<Rightarrow> mismatch tag ([fst x], snd x)"

text {* parses the first child *}
definition xml_take :: "'a xmlt2 \<Rightarrow> ('a \<Rightarrow> 'b xmlst) \<Rightarrow> 'b xmlst"
where "xml_take p1 p2 x \<equiv>
  case x of ([],rest) \<Rightarrow> (
    (* Only for accumulating expected tags. *)
    case p1 (XML [] [] [], rest) of Inr a \<Rightarrow> Inl (Fatal ''unexpected'')
    | Inl err \<Rightarrow> Inl err
  )
  | (x#xs,atts,flag,cands,rest) \<Rightarrow> (
      case p1 (x,atts,flag,cands,rest) of
        Inr a \<Rightarrow> p2 a (xs,atts,False,[],rest) (*If one child is parsed, then later mismatch is not allowed *)
      | Inl err \<Rightarrow>  Inl err
  )"

definition xml_take_text :: "(string \<Rightarrow> 'a xmlst) \<Rightarrow> 'a xmlst" where
  "xml_take_text p xs \<equiv>
   case xs of (XML_text text # xmls, s) \<Rightarrow> p text (xmls,s)
   | _ \<Rightarrow> xml_error ''expecting a text'' xs"

context begin

text {* A natural number parser which is proven correct: *}

definition nat_of_digit :: "char \<Rightarrow> nat option" where
  "nat_of_digit x \<equiv>
    if x = CHR ''0'' then Some 0
    else if x = CHR ''1'' then Some 1
    else if x = CHR ''2'' then Some 2
    else if x = CHR ''3'' then Some 3
    else if x = CHR ''4'' then Some 4
    else if x = CHR ''5'' then Some 5
    else if x = CHR ''6'' then Some 6
    else if x = CHR ''7'' then Some 7
    else if x = CHR ''8'' then Some 8
    else if x = CHR ''9'' then Some 9
    else None"

private fun nat_of_string_aux :: "nat \<Rightarrow> string \<Rightarrow> nat option"
where
  "nat_of_string_aux n [] = Some n" |
  "nat_of_string_aux n (d # s) = (nat_of_digit d \<bind> (\<lambda>m. nat_of_string_aux (10 * n + m) s))"

definition "nat_of_string s \<equiv>
  case if s = [] then None else nat_of_string_aux 0 s of
    None \<Rightarrow> Inl (''cannot convert \"'' @ s @ ''\" to a number'')
  | Some n \<Rightarrow> Inr n"

private lemma nat_of_string_aux_snoc:
  "nat_of_string_aux n (s @ [c]) =
   do { l \<leftarrow> nat_of_string_aux n s; m \<leftarrow> nat_of_digit c; Some (10 * l + m) }"
  by (induct s arbitrary:n, auto)

private lemma nat_of_string_aux_digit:
  assumes m10: "m < 10"
  shows "nat_of_string_aux n (s @ string_of_digit m) =
    do { l \<leftarrow> nat_of_string_aux n s; Some (10 * l + m) }"
proof-
 (* case analysis of 10 cases *)
  { note numeral_eq_Suc[simp]
    note m10[simplified]
  }
  from this[unfolded less_Suc_eq, folded numeral_1_eq_Suc_0, unfolded Suc_numeral, simplified]
  show ?thesis by (auto simp add: nat_of_digit_def nat_of_string_aux_snoc)
qed

declare string_of_digit.simps[simp del]

private lemmas shows_move = showsp_nat_append[of 0 _ "[]",simplified, folded shows_prec_nat_def]

private lemma nat_of_string_aux_show: "nat_of_string_aux 0 (show m) = Some m"
proof (induct m rule:less_induct)
  case IH: (less m)
  show ?case proof (cases "m < 10")
    case m10: True
    show ?thesis
      apply (unfold shows_prec_nat_def)
      apply (subst showsp_nat.simps)
      using m10 nat_of_string_aux_digit[OF m10, of 0 "[]"]
      apply (auto simp add:shows_string_def nat_of_string_def string_of_digit.simps)
      done
  next
    case m: False
    then have "m div 10 < m" by auto
    note IH = IH[OF this]
    show ?thesis apply (unfold shows_prec_nat_def, subst showsp_nat.simps)
      using m apply (auto simp: shows_prec_nat_def[symmetric] shows_string_def)
      apply (subst shows_move)
      using nat_of_string_aux_digit[OF semiring_numeral_div_class.pos_mod_bound,simplified]
      using m IH by (auto simp: nat_of_string_def)
  qed
qed

lemma fixes m :: nat shows show_nonemp: "show m \<noteq> []"
  apply (unfold shows_prec_nat_def)
  apply (subst showsp_nat.simps)
  apply (fold shows_prec_nat_def)
  apply (unfold o_def)
  apply (subst shows_move)
  apply (auto simp: shows_string_def string_of_digit.simps)
  done

lemma nat_of_string_show[simp]: "nat_of_string (show m) = Inr m"
  using nat_of_string_aux_show by (auto simp: nat_of_string_def show_nonemp)

end

fun safe_head where "safe_head [] = None" | "safe_head (x#xs) = Some x"

definition int_of_string :: "string \<Rightarrow> string + int"
where "int_of_string s \<equiv>
    if safe_head s = Some (CHR ''-'') then do { n \<leftarrow> nat_of_string (tl s); Inr (- int n) }
    else do { n \<leftarrow> nat_of_string s; Inr (int n) }"


definition xml_take_int :: "(int \<Rightarrow> 'a xmlst) \<Rightarrow> 'a xmlst" where
"xml_take_int p xs \<equiv>
  case xs of (XML_text text # xmls, s) \<Rightarrow>
    (case int_of_string text of Inl x \<Rightarrow> xml_error x xs | Inr n \<Rightarrow> p n (xmls,s))
  | _ \<Rightarrow> xml_error (''expecting an integer'') xs"

definition xml_take_nat :: "(nat \<Rightarrow> 'a xmlst) \<Rightarrow> 'a xmlst" where
  "xml_take_nat p xs \<equiv>
  case xs of (XML_text text # xmls, s) \<Rightarrow>
    (case nat_of_string text of Inl x \<Rightarrow> xml_error x xs | Inr n \<Rightarrow> p n (xmls,s))
  | _ \<Rightarrow> xml_error (''expecting a number'') xs"

definition xml_leaf where
"xml_leaf tag ret \<equiv> xml_do tag (xml_return ret)"

definition xml_text :: "tag \<Rightarrow> string xmlt2" where
"xml_text tag \<equiv> xml_do tag (xml_take_text xml_return)"

definition xml_int :: "tag \<Rightarrow> int xmlt2" where
"xml_int tag \<equiv> xml_do tag (xml_take_int xml_return)"

definition xml_nat :: "tag \<Rightarrow> nat xmlt2" where
  "xml_nat tag \<equiv> xml_do tag (xml_take_nat xml_return)"

definition bool_of_string :: "string \<Rightarrow> string + bool"
where
  "bool_of_string s \<equiv>
    if s = ''true'' then Inr True
    else if s = ''false'' then Inr False
    else Inl (''cannot convert \"'' @ s @ ''\" into Boolean'')"

definition xml_bool :: "tag \<Rightarrow> bool xmlt2"
where
  "xml_bool tag x \<equiv>
    case xml_text tag x of Inr str \<Rightarrow>
      ( case bool_of_string str of Inr b \<Rightarrow> Inr b
        | Inl err \<Rightarrow> xml_error err ([fst x], snd x)
      )
    | Inl err \<Rightarrow> Inl err"


(* Next one is a bit tricky to use. Invent better way... *)
definition xml_change :: "'a xmlt2 \<Rightarrow> ('a \<Rightarrow> 'b xmlst) \<Rightarrow> 'b xmlt2" where
  "xml_change p f x \<equiv>
    case p x of Inl err \<Rightarrow> Inl err
    | Inr a \<Rightarrow> case x of (_,rest) \<Rightarrow> f a ([],rest)"


text {*
  Parses the first child, if tag matches.
*}
definition xml_take_optional :: "'a xmlt2 \<Rightarrow> ('a option \<Rightarrow> 'b xmlst) \<Rightarrow> 'b xmlst"
where "xml_take_optional p1 p2 xs \<equiv>
  case xs of ([],_) \<Rightarrow> p2 None xs
  | (xml # xmls, atts, allow, cands, rest) \<Rightarrow> (
    case p1 (xml, atts, True, cands, rest) of Inr a \<Rightarrow> p2 (Some a) (xmls, atts, False, [], rest)
    | Inl (TagMismatch cands1) \<Rightarrow> p2 None (xml#xmls, atts, allow, cands1, rest) (* TagMismatch is allowed *)
    | Inl err \<Rightarrow> Inl err
  )"

definition xml_take_default :: "'a \<Rightarrow> 'a xmlt2 \<Rightarrow> ('a \<Rightarrow> 'b xmlst) \<Rightarrow> 'b xmlst"
where "xml_take_default a p1 p2 xs \<equiv>
  case xs of ([],_) \<Rightarrow> p2 a xs
  | (xml # xmls, atts, allow, cands, rest) \<Rightarrow> (
    case p1 (xml, atts, True, cands, rest) of Inr a \<Rightarrow> p2 a (xmls, atts, False, [], rest)
    | Inl (TagMismatch cands1) \<Rightarrow> p2 a (xml#xmls, atts, allow, cands1, rest)  (* TagMismatch is allowed *)
    | Inl err \<Rightarrow> Inl err
  )"

text {* Take first children, as many as tag matches. *}

fun xml_take_many_sub :: "'a list \<Rightarrow> nat \<Rightarrow> enat \<Rightarrow> 'a xmlt2 \<Rightarrow> ('a list \<Rightarrow> 'b xmlst) \<Rightarrow> 'b xmlst" where
  "xml_take_many_sub acc minOccurs maxOccurs p1 p2 ([], atts, allow, rest) = (
    if minOccurs = 0 then p2 (rev acc) ([], atts, allow, rest)
    else (* only for nice error log *)
      case p1 (XML [] [] [], atts, False, rest) of Inl err \<Rightarrow> Inl err
      | _ \<Rightarrow> Inl (Fatal ''unexpected'')
  )"
| "xml_take_many_sub acc minOccurs maxOccurs p1 p2 (xml # xmls, atts, allow, cands, rest) = (
      if maxOccurs = 0 then p2 (rev acc) (xml # xmls, atts, allow, cands, rest)
      else
        case p1 (xml, atts, minOccurs = 0, cands, rest)
        of Inr a \<Rightarrow> xml_take_many_sub (a # acc) (minOccurs-1) (maxOccurs-1) p1 p2 (xmls, atts, False, [], rest)
        | Inl (TagMismatch tags) \<Rightarrow> p2 (rev acc) (xml # xmls, atts, allow, cands, rest)
        | Inl err \<Rightarrow> Inl err
  )"

abbreviation xml_take_many where "xml_take_many \<equiv> xml_take_many_sub []"

fun pick_up where
"pick_up rest key [] = None"
| "pick_up rest key ((l,r)#s) = (if key = l then Some (r,rest@s) else pick_up ((l,r)#rest) key s)"

definition xml_take_attribute :: "string \<Rightarrow> (string \<Rightarrow> 'a xmlst) \<Rightarrow> 'a xmlst"
where "xml_take_attribute att p xs \<equiv>
  case xs of (xmls,atts,allow,cands,pos) \<Rightarrow> (
    case pick_up [] att atts of
      None \<Rightarrow> xml_error (''attribute \"'' @ att @''\" not found'') xs
    | Some(v,rest) \<Rightarrow> p v (xmls,rest,allow,cands,pos)
  )"

definition xml_take_attribute_optional :: "string \<Rightarrow> (string option \<Rightarrow> 'a xmlst) \<Rightarrow> 'a xmlst"
where "xml_take_attribute_optional att p xs \<equiv>
  case xs of (xmls,atts,info) \<Rightarrow> (
    case pick_up [] att atts of
      None \<Rightarrow> p None xs
    | Some(v,rest) \<Rightarrow> p (Some v) (xmls,rest,info)
  )"

definition xml_take_attribute_default :: "string \<Rightarrow> string \<Rightarrow> (string \<Rightarrow> 'a xmlst) \<Rightarrow> 'a xmlst"
where "xml_take_attribute_default def att p xs \<equiv>
  case xs of (xmls,atts,info) \<Rightarrow> (
    case pick_up [] att atts of
      None \<Rightarrow> p def xs
    | Some(v,rest) \<Rightarrow> p v (xmls,rest,info)
  )"

nonterminal xml_binds and xml_bind
syntax
  "_xml_block" :: "xml_binds \<Rightarrow> 'a" ("XMLdo {//(2  _)//}" [12] 1000)
  "_xml_take"  :: "pttrn \<Rightarrow> 'a \<Rightarrow> xml_bind" ("(2_ \<leftarrow>/ _)" 13)
  "_xml_take_att"  :: "pttrn \<Rightarrow> string \<Rightarrow> xml_bind" ("(2_ \<leftarrow>att/ _)" 13)
  "_xml_take_att_optional" :: "pttrn \<Rightarrow> string \<Rightarrow> xml_bind" ("(2_ \<leftarrow>att?/ _)" 13)
  "_xml_take_att_default" :: "pttrn \<Rightarrow> string \<Rightarrow> string \<Rightarrow> xml_bind" ("(2_ \<leftarrow>att[(_)]/ _)" 13)
  "_xml_take_optional" :: "pttrn \<Rightarrow> 'a \<Rightarrow> xml_bind" ("(2_ \<leftarrow>?/ _)" 13)
  "_xml_take_default" :: "pttrn \<Rightarrow> 'b \<Rightarrow> 'a \<Rightarrow> xml_bind" ("(2_ \<leftarrow>[(_)]/ _)" 13)
  "_xml_take_all" :: "pttrn \<Rightarrow> 'a \<Rightarrow> xml_bind" ("(2_ \<leftarrow>*/ _)" 13)
  "_xml_take_many" :: "pttrn \<Rightarrow> nat \<Rightarrow> enat \<Rightarrow> 'a \<Rightarrow> xml_bind" ("(2_ \<leftarrow>^{(_)..(_)}/ _)" 13)
  "_xml_let" :: "pttrn \<Rightarrow> 'a \<Rightarrow> xml_bind" ("(2let _ =/ _)" [1000, 13] 13)
  "_xml_final" :: "'a xmlst \<Rightarrow> xml_binds" ("_")
  "_xml_cons" :: "xml_bind \<Rightarrow> xml_binds \<Rightarrow> xml_binds" ("_;//_" [13, 12] 12)
  "_xml_do" :: "tag \<Rightarrow> xml_binds \<Rightarrow> 'a" ("XMLdo (_) {//(2 _)//}" [1000,12] 1000)

syntax (ASCII)
  "_xml_take" :: "pttrn \<Rightarrow> 'a \<Rightarrow> xml_bind" ("(2_ <-/ _)" 13)

translations
  "_xml_block (_xml_cons (_xml_take p x) (_xml_final e))"
    \<rightleftharpoons> "_xml_block (_xml_final (CONST xml_take x (\<lambda>p. e)))"
  "_xml_block (_xml_cons (_xml_take_att p x) (_xml_final e))"
    \<rightleftharpoons> "_xml_block (_xml_final (CONST xml_take_attribute x (\<lambda>p. e)))"
  "_xml_block (_xml_cons (_xml_take_att_optional p x) (_xml_final e))"
    \<rightleftharpoons> "_xml_block (_xml_final (CONST xml_take_attribute_optional x (\<lambda>p. e)))"
  "_xml_block (_xml_cons (_xml_take_att_default p d x) (_xml_final e))"
    \<rightleftharpoons> "_xml_block (_xml_final (CONST xml_take_attribute_default d x (\<lambda>p. e)))"
  "_xml_block (_xml_cons (_xml_take_optional p x) (_xml_final e))"
    \<rightleftharpoons> "_xml_block (_xml_final (CONST xml_take_optional x (\<lambda>p. e)))"
  "_xml_block (_xml_cons (_xml_take_default p d x) (_xml_final e))"
    \<rightleftharpoons> "_xml_block (_xml_final (CONST xml_take_default d x (\<lambda>p. e)))"
  "_xml_block (_xml_cons (_xml_take_all p x) (_xml_final e))"
    \<rightleftharpoons> "_xml_block (_xml_final (CONST xml_take_many 0 \<infinity> x (\<lambda>p. e)))"
  "_xml_block (_xml_cons (_xml_take_many p minOccurs maxOccurs x) (_xml_final e))"
    \<rightleftharpoons> "_xml_block (_xml_final (CONST xml_take_many minOccurs maxOccurs x (\<lambda>p. e)))"
  "_xml_block (_xml_cons (_xml_let p t) bs)"
    \<rightleftharpoons> "let p = t in _xml_block bs"
  "_xml_block (_xml_cons b (_xml_cons c cs))"
    \<rightleftharpoons> "_xml_block (_xml_cons b (_xml_final (_xml_block (_xml_cons c cs))))"
  "_xml_cons (_xml_let p t) (_xml_final s)"
    \<rightleftharpoons> "_xml_final (let p = t in s)"
  "_xml_block (_xml_final e)" \<rightharpoonup> "e"
  "_xml_do t e" \<rightleftharpoons> "CONST xml_do t (_xml_block e)"


definition xmlt_of_xmlt2 :: "'a xmlt2 \<Rightarrow> 'a xmlt" where
 "xmlt_of_xmlt2 p xml \<equiv>
  case (xml_take p xml_return) ([xml],[],False,[],[''??''])
  of Inl (Fatal str) \<Rightarrow> Left str
  | Inr a \<Rightarrow> Right a"

definition string2xml :: "string \<Rightarrow> string + xml" where
 "string2xml str \<equiv>
  case (do {
      update_tokens remove_comments;
      xml \<leftarrow> parse_node;
      eoi;
      Parser_Monad.return xml
    }) str
  of Inl err \<Rightarrow> Inl err | Inr (xml,_) \<Rightarrow> Inr xml"

definition parse_xml :: "'a xmlt2 \<Rightarrow> xml \<Rightarrow> string + 'a"
where "parse_xml p xml \<equiv>
    case xml_take p xml_return ([xml],[],False,[],[]) of Inr x \<Rightarrow> Inr x
    | Inl (Fatal x) \<Rightarrow> Inl x"

(*BEGIN: special chars*)
subsection \<open>Handling of special characters in text\<close>

definition "special_map = map_of [
  (''quot'', ''\"''), (''#34'', ''\"''), (*double quotation mark*)
  (''amp'', ''&''), (''#38'', ''&''), (*ampersand*)
  (''apos'', [CHR 0x27]), (''#39'', [CHR 0x27]), (*single quotes*)
  (''lt'', ''<''), (''#60'', ''<''), (*less-than sign*)
  (''gt'', ''>''), (''#62'', ''>'') (*greater-than sign*)
]"

fun extract_special
  where
    "extract_special acc [] = None"
  | "extract_special acc (x # xs) =
    (if x = CHR '';'' then map_option (\<lambda>s. (s, xs)) (special_map (rev acc))
    else extract_special (x#acc) xs)"

lemma extract_special_length [termination_simp]:
  assumes "extract_special acc xs = Some (y, ys)"
  shows "length ys < length xs"
  using assms by (induct acc xs rule: extract_special.induct) (auto split: if_splits)

fun normalize_special
  where
    "normalize_special [] = []"
  | "normalize_special (x # xs) =
      (if x = CHR ''&'' then
        (case extract_special [] xs of
          None \<Rightarrow> ''&'' @ normalize_special xs
        | Some (spec, ys) \<Rightarrow> spec @ normalize_special ys)
      else x # normalize_special xs)"

fun map_xml_text :: "(string \<Rightarrow> string) \<Rightarrow> xml \<Rightarrow> xml"
  where
    "map_xml_text f (XML t as cs) = XML t as (map (map_xml_text f) cs)"
  | "map_xml_text f (XML_text txt) = XML_text (f txt)"
(*END: special chars*)

definition parse_xmlfile :: "'a xmlt2 \<Rightarrow> string \<Rightarrow> string + 'a"
where
  "parse_xmlfile p str \<equiv> case Xml.doc_of_string str of
    Inl err \<Rightarrow> Inl err
  | Inr (XMLDOC header xml) \<Rightarrow> parse_xml p (map_xml_text normalize_special xml)"

definition parse_xml_string :: "'a xmlt2 \<Rightarrow> string \<Rightarrow> string + 'a"
where "parse_xml_string p str \<equiv> case string2xml str of
    Inl err \<Rightarrow> Inl err
  | Inr xml \<Rightarrow> parse_xml p xml"



subsection {* For Terminating Parsers *}

(*TODO: replace the default size of xml *)
primrec size_xml
  where "size_xml (XML_text str) = size str"
    |   "size_xml (XML tag atts xmls) = 1 + size tag + (\<Sum>xml \<leftarrow> xmls. size_xml xml)"

abbreviation "size_xml_state \<equiv> size_xml \<circ> fst"
abbreviation "size_xmls_state x \<equiv> (\<Sum>xml \<leftarrow> fst x. size_xml xml)"

(* termination_simp? *)
lemma size_xml_nth [dest]: "i < length xmls \<Longrightarrow> size_xml (xmls!i) \<le> sum_list (map size_xml xmls)"
  using elem_le_sum_list[of _ "map Xmlt2.size_xml _", unfolded length_map] by auto

lemma xml_or_cong[fundef_cong]:
  assumes "\<And>info. p (fst x, info) = p' (fst x, info)"
      and "\<And>info. q (fst x, info) = q' (fst x, info)"
      and "x = x'"
  shows "(p XMLor q) x = (p' XMLor q') x'"
  using assms
  by (cases x, auto simp: xml_or_def intro!: Option.bind_cong ext split:sum.split xml_error.split)

lemma xml_do_cong[fundef_cong]:
  fixes p :: "'a xmlst"
  assumes "\<And>atts xmls info. fst x = XML tag atts xmls \<Longrightarrow> p (xmls,atts,info) = p' (xmls,atts,info)"
      and "x = x'"
  shows "xml_do tag p x = xml_do tag p' x'"
  using assms by (cases x, auto simp: xml_do_def split: xml.split)

lemma xml_take_cong[fundef_cong]:
  fixes p :: "'a xmlt2" and q :: "'a \<Rightarrow> 'b xmlst"
  assumes "\<And>a as info. fst x = a#as \<Longrightarrow> p (a, info) = p' (a, info)"
      and "\<And>a as ret info info'. x' = (a#as,info) \<Longrightarrow> q ret (as, info') = q' ret (as, info')"
      and "\<And>info. p (XML [] [] [], info) = p' (XML [] [] [], info)"
      and "x = x'"
  shows "xml_take p q x = xml_take p' q' x'"
  using assms by (cases x, auto simp: xml_take_def intro!: Option.bind_cong split: list.split sum.split)

lemma xml_take_many_cong[fundef_cong]:
  fixes p :: "'a xmlt2" and q :: "'a list \<Rightarrow> 'b xmlst"
  assumes p: "\<And>n info. n < length (fst x) \<Longrightarrow> p (fst x' ! n, info) = p' (fst x' ! n, info)"
      and err: "\<And>info. p (XML [] [] [], info) = p' (XML [] [] [], info)"
      and q: "\<And>ret n info. q ret (drop n (fst x'), info) = q' ret (drop n (fst x'), info)"
      and xx': "x = x'"
  shows "xml_take_many_sub ret minOccurs maxOccurs p q x = xml_take_many_sub ret minOccurs maxOccurs p' q' x'"
proof-
  obtain as b where x: "x = (as,b)" by (cases x, auto)
  show ?thesis
  proof (insert p q, fold xx', unfold x, induct as arbitrary: b minOccurs maxOccurs ret)
    case Nil
    with err show ?case by (cases b, auto intro!: Option.bind_cong)
  next
    case (Cons a as)
    from Cons(2,3)[where n=0] Cons(2,3)[where n="Suc n" for n]
    show ?case by (cases b, auto intro!: Option.bind_cong Cons(1) split: sum.split xml_error.split)
  qed
qed

lemma xml_take_optional_cong[fundef_cong]:
  fixes p :: "'a xmlt2" and q :: "'a option \<Rightarrow> 'b xmlst"
  assumes "\<And>a as info. fst x = a # as \<Longrightarrow> p (a, info) = p' (a, info)"
      and "\<And>a as ret info. fst x = a # as \<Longrightarrow> q (Some ret) (as, info) = q' (Some ret) (as, info)"
      and "\<And>info. q None (fst x', info) = q' None (fst x', info)"
      and xx': "x = x'"
  shows "xml_take_optional p q x = xml_take_optional p' q' x'"
  using assms by (cases x', auto simp: xml_take_optional_def split: list.split sum.split xml_error.split intro!: Option.bind_cong)

lemma xml_take_default_cong[fundef_cong]:
  fixes p :: "'a xmlt2" and q :: "'a \<Rightarrow> 'b xmlst"
  assumes "\<And>a as info. fst x = a # as \<Longrightarrow> p (a, info) = p' (a, info)"
      and "\<And>a as ret info. fst x = a # as \<Longrightarrow> q ret (as, info) = q' ret (as, info)"
      and "\<And>info. q d (fst x', info) = q' d (fst x', info)"
      and xx': "x = x'"
  shows "xml_take_default d p q x = xml_take_default d p' q' x'"
  using assms by (cases x', auto simp: xml_take_default_def split: list.split sum.split xml_error.split intro!: Option.bind_cong)


lemma xml_change_cong[fundef_cong]:
  assumes "x = x'"
      and "p x' = p' x'"
      and "\<And>ret y. p x = Inr ret \<Longrightarrow> q ret y = q' ret y"
  shows "xml_change p q x = xml_change p' q' x'"
  using assms by (cases x', auto simp: xml_change_def split: option.split sum.split intro!: Option.bind_cong)

text {* Some fundef_cong lemmas, can't they be automatic? *}

lemma if_cong_applied[fundef_cong]:
"b = c \<Longrightarrow>
(c \<Longrightarrow> x z = u w) \<Longrightarrow>
(\<not> c \<Longrightarrow> y z = v w) \<Longrightarrow>
z = w \<Longrightarrow>
(if b then x else y) z =
(if c then u else v) w"
by auto

lemma option_case_cong[fundef_cong]:
    "option = option' \<Longrightarrow>
    (option' = None \<Longrightarrow> f1 z = g1 w) \<Longrightarrow>
    (\<And>x2. option' = Some x2 \<Longrightarrow> f2 x2 z = g2 x2 w) \<Longrightarrow>
    z = w \<Longrightarrow>
    (case option of None \<Rightarrow> f1 | Some x2 \<Rightarrow> f2 x2) z = (case option' of None \<Rightarrow> g1 | Some x2 \<Rightarrow> g2 x2) w"
by (cases option, auto)

lemma sum_case_cong[fundef_cong]:
  "s = sum' \<Longrightarrow>
  (\<And>x1. sum' = Inl x1 \<Longrightarrow> f1 x1 z = g1 x1 w) \<Longrightarrow>
  (\<And>x2. sum' = Inr x2 \<Longrightarrow> f2 x2 z = g2 x2 w) \<Longrightarrow>
  z = w \<Longrightarrow>
  (case s of Inl x1 \<Rightarrow> f1 x1 | Inr x2 \<Rightarrow> f2 x2) z = (case sum' of Inl x1 \<Rightarrow> g1 x1 | Inr x2 \<Rightarrow> g2 x2) w"
by (cases s, auto)

lemma prod_case_cong[fundef_cong]: "p = prod' \<Longrightarrow>
  (\<And>x1 x2. prod' = (x1, x2) \<Longrightarrow> f x1 x2 z = g x1 x2 w) \<Longrightarrow>
  (case p of (x1, x2) \<Rightarrow> f x1 x2) z = (case prod' of (x1, x2) \<Rightarrow> g x1 x2) w"
  by (auto split: prod.split)

end
