(*
Author:  Bertram Felgenhauer <bertram.felgenhauer@uibk.ac.at> (2015-2017)
Author:  Franziska Rapp <franziska.rapp@uibk.ac.at> (2015-2017)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>Miscellaneous\<close>

text \<open>This theory contains several trivial extensions to existing IsaFoR theories.\<close>
(* TODO: move things elsewhere *)

theory LS_Extras
  imports TA.Multihole_Context QTRS.Renaming QTRS.Unification_More
begin

subsection \<open>Bounded duplication\<close>

lemma refl_inv_image:
  "refl R \<Longrightarrow> refl (inv_image R f)"
  by (simp add: inv_image_def refl_on_def)

lemma image_mset_Some_None_zero [simp]:
  "count (image_mset Some M) None = 0"
  by (induct M) auto

text \<open>IsaFoR does not seem to define non-duplicating TRSs yet.\<close>

definition non_duplicating :: "('f, 'v) trs \<Rightarrow> bool" where
  "non_duplicating R \<equiv> \<forall>l r. (l, r) \<in> R \<longrightarrow> vars_term_ms r \<le># vars_term_ms l"

text \<open>Bounded duplication @{cite \<open>Definition 4.2\<close> FMZvO15}. Note: @{term "undefined"} is an arbitrary fixed value.\<close>

definition bounded_duplicating :: "('f, 'v) trs \<Rightarrow> bool" where
  "bounded_duplicating R \<equiv> SN_rel (rstep {(Fun None [Var undefined], Var undefined)}) (rstep (map_funs_trs Some R))"

text \<open>Non-duplicating TRSs are bounded duplicating @{cite \<open>Lemma 4.4\<close> FMZvO15}.\<close>

lemma non_duplicating_imp_bounded_duplicating:
  assumes nd: "non_duplicating R"
  shows "bounded_duplicating R"
proof -
  let ?gt = "inv_image {(n, m) . n > m} (\<lambda>t. count (funs_term_ms t) None)"
  let ?ss = "rstep {(Fun None [Var undefined], Var undefined)}"
  let ?ge = "inv_image {(n, m) . n \<ge> m} (\<lambda>t. count (funs_term_ms t) None)"
  let ?rs = "rstep (map_funs_trs Some R)"
  have "?ss \<subseteq> ?gt"
  proof
    fix s t
    assume "(s, t) \<in> ?ss"
    then have "count (funs_term_ms s) None > count (funs_term_ms t) None"
    proof
      fix C \<sigma> l r
      assume rl: "(l, r) \<in> {(Fun None [Var undefined], Var undefined)}"
        and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      show "count (funs_term_ms s) None > count (funs_term_ms t) None" using rl
        by (auto simp: s t funs_term_ms_ctxt_apply funs_term_ms_subst_apply) 
    qed
    then show "(s, t) \<in> ?gt" by simp
  qed
  moreover have "?rs \<subseteq>  ?ge"
  proof
    fix s t
    assume "(s, t) \<in> ?rs"
    then have "count (funs_term_ms s) None \<ge> count (funs_term_ms t) None"
    proof
      fix C \<sigma> l r
      {
        fix l r
        assume "(l, r) \<in> map_funs_trs Some R"
        then have "vars_term_ms r \<le># vars_term_ms l \<and> count (funs_term_ms l) None = 0 \<and> count (funs_term_ms r) None = 0"
          by (auto simp: nd[unfolded non_duplicating_def] map_funs_trs.simps funs_term_ms_map_funs_term)
      } note * = this
      assume rl: "(l, r) \<in> map_funs_trs Some R"
        and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      from rl *[of l r] show "count (funs_term_ms s) None \<ge> count (funs_term_ms t) None"
        by (auto simp: s t funs_term_ms_ctxt_apply funs_term_ms_subst_apply)
           (metis image_mset_union mset_subset_eq_count subset_mset.le_iff_add sum_mset.union)
    qed
    then show "(s, t) \<in> ?ge" by simp
  qed
  moreover
  {
    have [simp]: "?ge\<^sup>* = ?ge"
      by (auto intro!: trans_refl_imp_rtrancl_id trans_inv_image refl_inv_image simp: refl_on_def trans_def)
    have *: "(?ge\<^sup>* O ?gt O ?ge\<^sup>*)\<inverse> \<subseteq> ?gt\<inverse>"
      by auto
    moreover have "wf (?gt\<inverse>)" by auto (auto simp: converse_def wf_less)
    ultimately have "SN_rel ?gt ?ge" unfolding SN_rel_on_def SN_iff_wf
      by (intro wf_subset[OF _ *]) auto
  }
  ultimately show ?thesis
    by (auto intro: SN_rel_mono simp: bounded_duplicating_def)
qed  

subsection \<open>Trivialities\<close>

lemma pos_diff_cons [simp]: "pos_diff (i <# p) (i <# q) = pos_diff p q"
  by (auto simp: pos_diff_def)

lemma max_list_append: "max_list (xs1 @ xs2) = max (max_list xs1) (max_list xs2)"
  by (induct xs1) auto

lemma max_list_bound:
  "max_list xs \<le> z \<longleftrightarrow> (\<forall>i < length xs. xs ! i \<le> z)"
  using less_Suc_eq_0_disj by (induct xs) auto

lemma max_list_bound_set:
  "max_list xs \<le> z \<longleftrightarrow> (\<forall>x \<in> set xs. x \<le> z)"
  using less_Suc_eq_0_disj by (induct xs) auto

lemma max_list_mono_concat:
  assumes "length xss = length yss" and "\<And>i. i < length xss \<Longrightarrow> max_list (xss ! i) \<le> max_list (yss ! i)"
  shows "max_list (concat xss) \<le> max_list (concat yss)"
  using assms
proof (induct yss arbitrary: xss)
  case (Cons ys yss) thus ?case by (cases xss) (force simp: max_list_append)+
qed auto

lemma max_list_mono_concat1:
  assumes "length xss = length ys" and "\<And>i. i < length xss \<Longrightarrow> max_list (xss ! i) \<le> ys ! i"
  shows "max_list (concat xss) \<le> max_list ys"
  using assms max_list_mono_concat[of xss "map (\<lambda>y. [y]) ys"] by auto

lemma take1:
  "take (Suc 0) xs = (if xs = [] then [] else [hd xs])"
  by (cases xs) auto

lemma sum_list_take':
  "i \<le> length (xs :: nat list) \<Longrightarrow> sum_list (take i xs) \<le> sum_list xs"
  by (induct xs arbitrary: i) (simp, case_tac i, auto)

lemma nth_equalityE:
  "xs = ys \<Longrightarrow> (length xs = length ys \<Longrightarrow> (\<And>i. i < length xs \<Longrightarrow> xs ! i = ys ! i) \<Longrightarrow> P) \<Longrightarrow> P"
  by simp

lemma finite_vars_mctxt [simp]: "finite (vars_mctxt C)"
  by (induct C) auto

lemma partition_by_of_zip:
  "length xs = sum_list zs \<Longrightarrow> length ys = sum_list zs \<Longrightarrow>
   partition_by (zip xs ys) zs = map (\<lambda>(x,y). zip x y) (zip (partition_by xs zs) (partition_by ys zs))"
  by (induct zs arbitrary: xs ys) (auto simp: take_zip drop_zip)

lemma mctxt_of_term_fill_holes':
  "num_holes C = length ts \<Longrightarrow> mctxt_of_term (fill_holes C ts) = fill_holes_mctxt C (map mctxt_of_term ts)"
  by (induct C ts rule: fill_holes_induct) auto

lemma vars_term_fill_holes':
  "num_holes C = length ts \<Longrightarrow> vars_term (fill_holes C ts) = \<Union>(vars_term ` set ts) \<union> vars_mctxt C"
proof (induct C ts rule: fill_holes_induct)
  case (MFun f Cs ts) thus ?case
    using UN_upt_len_conv[of "partition_holes ts Cs" "length Cs" "\<lambda>t. (\<Union>x\<in>set t. vars_term x)"]
    by (simp add: UN_Un_distrib UN_set_partition_by)
qed auto

lemma distinct_count_atmost_1':
  "distinct xs = (\<forall>a. count (mset xs) a \<le> 1)"
  unfolding distinct_count_atmost_1 using dual_order.antisym by fastforce

lemma nth_subset_concat:
  assumes "i < length xss" 
  shows "set (xss ! i) \<subseteq> set (concat xss)"
  by (metis assms concat_nth concat_nth_length in_set_idx nth_mem subsetI)

lemma subst_domain_subst_of:
  "subst_domain (subst_of xs) \<subseteq> set (map fst xs)"
proof (induct xs)
  case (Cons x xs)
  moreover have "subst_domain (subst (fst x) (snd x)) \<subseteq> set (map fst [x])" by simp
  ultimately show ?case
    using subst_domain_compose[of "subst_of xs" "subst (fst x) (snd x)"] by auto
qed simp

lemma subst_apply_mctxt_cong: "(\<forall>x. x \<in> vars_mctxt C \<longrightarrow> \<sigma> x = \<tau> x) \<Longrightarrow> C \<cdot>mc \<sigma> = C \<cdot>mc \<tau>"
  by (induct C) auto

lemma distinct_concat_unique_index:
  "distinct (concat xss) \<Longrightarrow> i < length xss \<Longrightarrow> x \<in> set (xss ! i) \<Longrightarrow> j < length xss \<Longrightarrow> x \<in> set (xss ! j) \<Longrightarrow> i = j"
proof (induct xss rule: List.rev_induct)
  case (snoc xs xss) thus ?case using nth_mem[of i xss] nth_mem[of j xss]
    by (cases "i < length xss"; cases "j < length xss") (auto simp: nth_append simp del: nth_mem)
qed auto

lemma list_update_concat:
  assumes "i < length xss" "j < length (xss ! i)"
  shows "concat (xss[i := (xss ! i)[j := x]]) = concat xss[sum_list (take i (map length xss)) + j := x]"
  using assms
proof (induct xss arbitrary: i)
  case (Cons xs xss) thus ?case
    by (cases i) (auto simp: list_update_append)
qed auto

lemma length_filter_sum_list:
  "length (filter p xs) = sum_list (map (\<lambda>x. if p x then 1 else 0) xs)"
  by (induct xs) auto

lemma poss_mctxt_mono:
  "C \<le> D \<Longrightarrow> poss_mctxt C \<subseteq> poss_mctxt D"
  by (induct C D rule: less_eq_mctxt_induct) force+

lemma poss_append_funposs:
  shows "p <#> q \<in> funposs t \<longleftrightarrow> p \<in> poss t \<and> q \<in> funposs (t |_ p)" (is "?L \<longleftrightarrow> ?R")
proof
  show "?L \<Longrightarrow> ?R" using funposs_imp_poss[of "p <#> q" t] funposs_fun_conv[of "p <#> q" t]
    poss_is_Fun_funposs[of q "t |_ p"] by auto
next
  show "?R \<Longrightarrow> ?L" using funposs_imp_poss[of q "t |_ p"] funposs_fun_conv[of q "t |_ p"]
    poss_is_Fun_funposs[of "p <#> q" t] by auto
qed

lemma mctxt_of_term_var_subst:
  "mctxt_of_term (t \<cdot> (Var \<circ> f)) = map_vars_mctxt f (mctxt_of_term t)"
  by (induct t) auto

lemma subst_apply_mctxt_map_vars_mctxt_conv:
  "C \<cdot>mc (Var \<circ> f) = map_vars_mctxt f C"
  by (induct C) auto

lemma map_vars_mctxt_mono:
  "C \<le> D \<Longrightarrow> map_vars_mctxt f C \<le> map_vars_mctxt f D"
  by (induct C D rule: less_eq_mctxt_induct) (auto intro: less_eq_mctxtI1)

lemma map_vars_mctxt_less_eq_decomp:
  assumes "C \<le> map_vars_mctxt f D"
  obtains C' where "map_vars_mctxt f C' = C" "C' \<le> D"
  using assms
proof (induct D arbitrary: C thesis)
  case (MVar x) show ?case using MVar(1)[of MHole] MVar(1)[of "MVar _"] MVar(2)
    by (auto elim: less_eq_mctxtE2 intro: less_eq_mctxtI1)
next
  case MHole show ?case using MHole(1)[of MHole] MHole(2) by (auto elim: less_eq_mctxtE2)
next
  case (MFun g Ds) note MFun' = MFun
  show ?case using MFun(3) unfolding map_vars_mctxt.simps
  proof (cases rule: less_eq_mctxtE2(3))
    case MHole then show ?thesis using MFun(2)[of MHole] by auto
  next
    case (MFun Cs)
    define Cs' where "Cs' = map (\<lambda>i. SOME Ci'. map_vars_mctxt f Ci' = Cs ! i \<and> Ci' \<le> Ds ! i) [0..<length Cs]"
    { fix i assume "i < length Cs"
      obtain Ci' where "map_vars_mctxt f Ci' = Cs ! i" "Ci' \<le> Ds ! i"
      using `i < length Cs` MFun MFun'(1)[OF nth_mem, of i] MFun'(3) by (auto elim!: less_eq_mctxtE2)
      then have "\<exists>Ci'. map_vars_mctxt f Ci' = Cs ! i \<and> Ci' \<le> Ds ! i" by blast
    }
    from someI_ex[OF this] have
      "length Cs = length Cs'" and "i < length Cs \<Longrightarrow> map_vars_mctxt f (Cs' ! i) = Cs ! i"
      "i < length Cs \<Longrightarrow> Cs' ! i \<le> Ds ! i" for i by (auto simp: Cs'_def)
    then show ?thesis using MFun(1,2) MFun'(3)
      by (auto intro!: MFun'(2)[of "MFun g Cs'"] nth_equalityI less_eq_mctxtI2 elim!: less_eq_mctxtE2)
  qed
qed

text \<open>see @{thm nrrstep_preserves_root} which proves this without the length\<close>
lemma nrrstep_preserves_root':
  assumes "(Fun f ss, t) \<in> nrrstep R"
  shows "\<exists>ts. t = Fun f ts \<and> length ss = length ts"
  using assms unfolding nrrstep_def rstep_r_p_s_def Let_def by auto

text \<open>see @{thm nrrsteps_preserve_root} which proves this without the length\<close>
lemma nrrsteps_preserve_root':
  assumes "(Fun f ss, t) \<in> (nrrstep R)\<^sup>*"
  shows "\<exists>ts. t = Fun f ts \<and> length ss = length ts"
  using assms by induct (auto dest: nrrstep_preserves_root')

lemma args_joinable_imp_joinable:
  assumes "length ss = length ts" "\<And>i. i < length ss \<Longrightarrow> (ss ! i, ts ! i) \<in> (rstep R)\<^sup>\<down>"
  shows "(Fun f ss, Fun f ts) \<in> (rstep R)\<^sup>\<down>"
proof -
  obtain u where "i < length ss \<Longrightarrow> (ss ! i, u i) \<in> (rstep R)\<^sup>* \<and> (ts ! i, u i) \<in> (rstep R)\<^sup>*" for i
    using joinD[OF assms(2)] by metis
  then show ?thesis using assms(1)
    by (intro joinI[of _ "Fun f (map u [0..<length ss])"] args_rsteps_imp_rsteps) auto
qed

instance option :: (infinite) infinite
  by standard (simp add: infinite_UNIV)

lemma finite_into_infinite:
  assumes "finite A" "infinite B"
  shows "\<exists>f. f ` A \<subseteq> B \<and> inj_on f A"
proof -
  from finite_imp_inj_to_nat_seg[OF assms(1)]
  obtain f :: "_ \<Rightarrow> nat" and n where "f ` A = {i. i < n}" "inj_on f A" by auto
  moreover from infinite_countable_subset[OF assms(2)]
  obtain g :: "nat \<Rightarrow> _" where "inj g" "range g \<subseteq> B" by auto
  ultimately show ?thesis by (auto simp: inj_on_def intro!: exI[of _ "g \<circ> f"])
qed

lemma finites_into_infinites:
  fixes f :: "'a \<Rightarrow> 'b set" and g :: "'a \<Rightarrow> 'c set"
  assumes "\<And>\<alpha> \<beta>. f \<alpha> \<inter> f \<beta> \<noteq> {} \<Longrightarrow> \<alpha> = \<beta>"
  and "\<And>\<alpha>. finite (f \<alpha>)"
  and "\<And>\<alpha> \<beta>. g \<alpha> \<inter> g \<beta> \<noteq> {} \<Longrightarrow> \<alpha> = \<beta>"
  and "\<And>\<alpha>. infinite (g \<alpha>)"
  shows "\<exists>h :: 'b \<Rightarrow> 'c. inj_on h (\<Union>\<alpha>. f \<alpha>) \<and> (\<forall>\<alpha>. h ` f \<alpha> \<subseteq> g \<alpha>)"
proof -
  from finite_into_infinite[OF assms(2,4)] have "\<exists>h. h ` f \<alpha> \<subseteq> g \<alpha> \<and> inj_on h (f \<alpha>)" for \<alpha> by blast
  then obtain h where h: "h \<alpha> ` f \<alpha> \<subseteq> g \<alpha> \<and> inj_on (h \<alpha>) (f \<alpha>)" for \<alpha> by metis
  have [simp]: "x \<in> f \<alpha> \<Longrightarrow> (THE \<alpha>. x \<in> f \<alpha>) = \<alpha>" for x \<alpha> using assms(1) by auto
  then show ?thesis using assms(1,3) h
    apply (intro exI[of _ "\<lambda>x. h (THE \<alpha>. x \<in> f \<alpha>) x"])
    apply (auto simp: inj_on_def image_def)
    by blast
qed

subsection \<open>Imbalance\<close>

definition refines :: "'a list \<Rightarrow> 'b list \<Rightarrow> bool" (infix "\<propto>" 55) where
  "ss \<propto> ts \<longleftrightarrow> length ts = length ss \<and> (\<forall>i j. i < length ss \<and> j < length ts \<and> ss ! i = ss ! j \<longrightarrow> ts ! i = ts ! j)"

lemma refines_refl:
  "ss \<propto> ss"
  by (auto simp: refines_def)

lemma refines_trans:
  "ss \<propto> ts \<Longrightarrow> ts \<propto> us \<Longrightarrow> ss \<propto> us"
  by (auto simp: refines_def)

lemma refines_map:
  "ss \<propto> map f ss"
  by (auto simp: refines_def)

lemma refines_imp_map:
  assumes "ss \<propto> ts"
  obtains f where "ts = map f ss"
proof -
  note * = assms[unfolded refines_def]
  let ?P = "\<lambda>s i. (i < length ss \<and> ss ! i = s)" let ?f = "\<lambda>s. ts ! (SOME i. ?P s i)"
  { fix i assume "i < length ss"
    then have "ts ! i = ?f (ss ! i)" using conjunct1[OF *]
      conjunct2[OF *, THEN spec[of _ i], THEN spec[of _ "SOME j. ?P (ss ! i) j"]] someI[of "?P (ss ! i)" i]
      by auto
  }
  then show ?thesis using conjunct1[OF *] by (auto intro!: that[of "?f"] nth_equalityI)
qed

definition imbalance :: "'a list \<Rightarrow> nat" where
  "imbalance ts = card (set ts)"

lemma imbalance_def':
  "imbalance xs = card { i. i < length xs \<and> (\<forall>j. j < length xs \<and> xs ! i = xs ! j \<longrightarrow> i \<le> j) }"
proof (induct xs rule: List.rev_induct)
  case (snoc x xs)
  have "{ i. i < length (xs @ [x]) \<and> (\<forall>j. j < length (xs @ [x]) \<and> (xs @ [x]) ! i = (xs @ [x]) ! j \<longrightarrow> i \<le> j) } =
        { i. i < length xs \<and> (\<forall>j. j < length xs \<and> xs ! i = xs ! j \<longrightarrow> i \<le> j) } \<union> { i. i = length xs \<and> x \<notin> set xs }"
    by (simp add: set_eq_iff less_Suc_eq dnf imp_conjL nth_append)
      (metis cancel_comm_monoid_add_class.diff_cancel in_set_conv_nth nth_Cons_0)
  then show ?case using snoc by (simp add: imbalance_def card_insert_if)
qed (auto simp: imbalance_def)

lemma imbalance_mono: "set ss \<subseteq> set ts \<Longrightarrow> imbalance ss \<le> imbalance ts"
  by (simp add: imbalance_def card_mono)

lemma refines_imbalance_mono:
  "ss \<propto> ts \<Longrightarrow> imbalance ss \<ge> imbalance ts"
  (* apply (auto simp: refines_def imbalance_def' intro: card_mono) (* desired proof *) *)
  unfolding refines_def imbalance_def' by (intro card_mono) (simp_all add: Collect_mono)

lemma refines_imbalance_strict_mono:
  "ss \<propto> ts \<Longrightarrow> \<not> ts \<propto> ss \<Longrightarrow> imbalance ss > imbalance ts"
  unfolding refines_def imbalance_def'
proof (intro psubset_card_mono psubsetI, goal_cases)
  case 3 obtain i j where ij: "i < length ss" "j < length ss" "ts ! i = ts ! j" "ss ! i \<noteq> ss ! j"
    using conjunct1[OF 3(1)] 3(2) by auto
  define f where "f \<equiv> \<lambda>i. SOME j. j < length ss \<and> ss ! j = ss ! i \<and> (\<forall>i. ss ! j = ss ! i \<longrightarrow> j \<le> i)"
  let ?ssi = "{i. i < length ss \<and> (\<forall>j. j < length ss \<and> ss ! i = ss ! j \<longrightarrow> i \<le> j)}"
  let ?tsi = "{i. i < length ts \<and> (\<forall>j. j < length ts \<and> ts ! i = ts ! j \<longrightarrow> i \<le> j)}"
  note f_def = fun_cong[OF f_def[unfolded atomize_eq]]
  { fix i assume "i < length ss"
    then have "\<exists>i'. i' < length ss \<and> ss ! i' = ss ! i \<and> (\<forall>i. ss ! i' = ss ! i \<longrightarrow> i' \<le> i)"
    proof (induct i rule: less_induct)
      case (less i) then show ?case
        by (cases "i < length ss \<and> ss ! i = ss ! i \<and> (\<forall>j. ss ! i = ss ! j \<longrightarrow> i \<le> j)") auto
    qed
    then have "f i < length ss" "ss ! f i = ss ! i" "\<And>j. ss ! f i = ss ! j \<Longrightarrow> f i \<le> j"
    using someI[of "\<lambda>j. j < length ss \<and> ss ! j = ss ! i \<and> (\<forall>i. ss ! j = ss ! i \<longrightarrow> j \<le> i)", folded f_def]
      by auto
    then have "f i < length ss" "ss ! f i = ss ! i" "f i \<in> ?ssi" "f i \<in> ?tsi \<Longrightarrow> ts ! f i = ts ! i"
      using 3(1) by (auto simp: `i < length ss`)
  } note * = this[OF ij(1)] this[OF ij(2)]
  with ij(1,2,3,4) 3(1)[THEN conjunct1]
    3(1)[THEN conjunct2, rule_format, of i "f i"]
    3(1)[THEN conjunct2, rule_format, of j "f j"]
  have "f i \<in> ?ssi" "f j \<in> ?ssi" "ts ! f i = ts ! f j" "f i \<noteq> f j" by metis+
  moreover from this(3,4) have "f i \<notin> ?tsi \<or> f j \<notin> ?tsi" using *(4,8) ij(3)
    by (metis (mono_tags, lifting) dual_order.antisym mem_Collect_eq)
  ultimately show ?case by argo
qed (simp_all add: Collect_mono)

lemma refines_take:
  "ss \<propto> ts \<Longrightarrow> take n ss \<propto> take n ts"
  unfolding refines_def by (intro conjI impI allI; elim conjE) (simp_all add: refines_def)

lemma refines_drop:
  "ss \<propto> ts \<Longrightarrow> drop n ss \<propto> drop n ts"
  unfolding refines_def
proof ((intro conjI impI allI; elim conjE), goal_cases)
  case (2 i j) show ?case using 2(1,3-) 2(2)[rule_format, of "i + n" "j + n"]
    by (auto simp: less_diff_eq less_diff_conv ac_simps)
qed simp

subsection \<open>Abstract Rewriting\<close>

lemma join_finite:
  assumes "CR R" "finite X" "\<And>x y. x \<in> X \<Longrightarrow> y \<in> X \<Longrightarrow> (x, y) \<in> R\<^sup>\<leftrightarrow>\<^sup>*"
  obtains z where "\<And>x. x \<in> X \<Longrightarrow> (x, z) \<in> R\<^sup>*"
  using assms(2,1,3)
proof (induct X arbitrary: thesis)
  case (insert x X)
  then show ?case
  proof (cases "X = {}")
    case False
    then obtain x' where "x' \<in> X" by auto
    obtain z where *: "x' \<in> X \<Longrightarrow> (x', z) \<in> R\<^sup>*" for x' using insert by (metis insert_iff)
    from this[of x'] `x' \<in> X` insert(6)[of x x'] have "(x, z) \<in> R\<^sup>\<leftrightarrow>\<^sup>*"
      by (metis insert_iff converse_rtrancl_into_rtrancl conversionI' conversion_rtrancl)
    with `CR R` obtain z' where "(x, z') \<in> R\<^sup>*" "(z, z') \<in> R\<^sup>*" by (auto simp: CR_iff_conversion_imp_join)
    then show ?thesis by (auto intro!: insert(4)[of z'] dest: *)
  qed auto
qed simp

lemma balance_sequence:
  assumes "CR R"
  obtains (ts) ts where "length ss = length ts"
    "\<And>i. i < length ss \<Longrightarrow> (ss ! i, ts ! i) \<in> R\<^sup>*"
    "\<And>i j. i < length ss \<Longrightarrow> j < length ss \<Longrightarrow> (ss ! i, ss ! j) \<in> R\<^sup>\<leftrightarrow>\<^sup>* \<Longrightarrow> ts ! i = ts ! j"
proof -
  define f where "f s \<equiv> SOME u. \<forall>t \<in> { t \<in> set ss. (s, t) \<in> R\<^sup>\<leftrightarrow>\<^sup>* }. (t, u) \<in> R\<^sup>* " for s
  { fix i assume "i < length ss"
    have "\<forall>t \<in> { t \<in> set ss. (ss ! i, t) \<in> R\<^sup>\<leftrightarrow>\<^sup>* }. (t, f (ss ! i)) \<in> R\<^sup>*"
      unfolding f_def using join_finite[OF `CR R`, of "{ t \<in> set ss. (ss ! i, t) \<in> R\<^sup>\<leftrightarrow>\<^sup>* }"]
        by (rule someI_ex) (auto, metis conversion_inv conversion_rtrancl rtrancl_trans)
    with `i < length ss` have "(ss ! i, f (ss ! i)) \<in> R\<^sup>*" by auto
  } note [intro] = this
  moreover {
    fix i j assume "i < length ss" "j < length ss" "(ss ! i, ss ! j) \<in> R\<^sup>\<leftrightarrow>\<^sup>*"
    then have "{ t \<in> set ss. (ss ! i, t) \<in> R\<^sup>\<leftrightarrow>\<^sup>* } = { t \<in> set ss. (ss ! j, t) \<in> R\<^sup>\<leftrightarrow>\<^sup>* }"
      by auto (metis conversion_inv conversion_rtrancl rtrancl_trans)+
    then have "f (ss ! i) = f (ss ! j)" by (auto simp: f_def)
  } note [intro] = this
  show ?thesis by (auto intro!: ts[of "map f ss"])
qed

lemma balance_sequences:
  assumes "CR R" and [simp]: "length ts = length ss" "length us = length ss" and
    p: "\<And>i. i < length ss \<Longrightarrow> (ss ! i, ts ! i) \<in> R\<^sup>*" "\<And>i. i < length ss \<Longrightarrow> (ss ! i, us ! i) \<in> R\<^sup>*"
  obtains (vs) vs where
    "length vs = length ss"
    "\<And>i. i < length ss \<Longrightarrow> (ts ! i, vs ! i) \<in> R\<^sup>*" "\<And>i. i < length ss \<Longrightarrow> (us ! i, vs ! i) \<in> R\<^sup>*"
    "refines ts vs" "refines us vs"
proof -
  from balance_sequence[OF `CR R`, of "ts @ us"]
  obtain vs where l: "length (ts @ us) = length vs" and
    r: "\<And>i. i < length (ts @ us) \<Longrightarrow> ((ts @ us) ! i, vs ! i) \<in> R\<^sup>*" and
    e: "\<And>i j. i < length (ts @ us) \<Longrightarrow> j < length (ts @ us) \<Longrightarrow> ((ts @ us) ! i, (ts @ us) ! j) \<in> R\<^sup>\<leftrightarrow>\<^sup>* \<Longrightarrow> vs ! i = vs ! j"
    by blast
  { fix i assume *: "i < length ss"
    from * have "(ts ! i, us ! i) \<in> R\<^sup>\<leftrightarrow>\<^sup>*" using p[of i]
      by (metis CR_imp_conversionIff_join assms(1) conversionI' conversion_rtrancl joinI_right rtrancl_trans)
    with * have "vs ! (length ss + i) = vs ! i"
      using e[of i "length ss + i"] by (auto simp: nth_append)
  } note lp[simp] = this
  show ?thesis
  proof (intro vs[of "take (length ss) vs"], goal_cases)
    case (2 i) then show ?case using l r[of i] by (auto simp: nth_append)
  next
    case (3 i) with 3 show ?case using l r[of "length ss + i"] by (auto simp: nth_append)
  next
    case 4
    then show ?case using l by (auto simp: refines_def nth_append intro!: e)
  next
    case 5
    { fix i j assume "i < length ss" "j < length ss" "(us ! i, us ! j) \<in> R\<^sup>\<leftrightarrow>\<^sup>*"
      then have "vs ! i = vs ! j"
        using e[of "length ss + i" "length ss + j"] by (simp add: nth_append)
    }
    then show ?case using l by (auto simp: refines_def nth_append)
  qed (auto simp: l[symmetric])
qed

lemma rtrancl_on_iff_rtrancl_restr:
  assumes "\<And>x y. x \<in> A \<Longrightarrow> (x, y) \<in> R \<Longrightarrow> y \<in> A"
    "(x, y) \<in> R\<^sup>*" "x \<in> A" 
  shows "(x, y) \<in> (Restr R A)\<^sup>* \<and> y \<in> A"
  using assms(2,3)
proof (induct y rule: rtrancl_induct)
  case (step y z) then show ?case by (auto intro!: rtrancl_into_rtrancl[of _ y _ z] simp: assms(1))
qed auto

lemma CR_on_iff_CR_Restr:
  assumes "\<And>x y. x \<in> A \<Longrightarrow> (x, y) \<in> R \<Longrightarrow> y \<in> A"
  shows "CR_on R A \<longleftrightarrow> CR (Restr R A)"
proof ((standard; standard), goal_cases)
  let ?R' = "Restr R A"
  note * = rtrancl_on_iff_rtrancl_restr[OF assms]
  have *: "(s, t) \<in> R\<^sup>* \<Longrightarrow> s \<in> A \<Longrightarrow> (s, t) \<in> ?R'\<^sup>* \<and> t \<in> A" for s t
    using assms rtrancl_on_iff_rtrancl_restr by metis
  {
    case (1 s t u) then show ?case
    proof (cases "s \<in> A")
      case True
      with 1(3,4) have "(s, t) \<in> R\<^sup>*" "t \<in> A" "(s, u) \<in> R\<^sup>*" "u \<in> A"
        using *[of s t] *[of s u] rtrancl_mono[of ?R' R] by auto
      then show ?thesis
        using True 1(1)[unfolded CR_on_def, rule_format, of s t u] by (auto dest!: joinD *)
    next
      case False
      then have "s = t" "s = u" using 1 by (metis IntD2 converse_rtranclE mem_Sigma_iff)+
      then show ?thesis by auto
    qed
  next
    case (2 s t u)
    obtain v where "(t, v) \<in> ?R'\<^sup>*" "(u, v) \<in> ?R'\<^sup>*" using *[OF 2(3,2)] *[OF 2(4,2)] 2(1) by auto
    then show ?case using rtrancl_mono[of ?R' R] by auto
  }
qed

subsection \<open>Bijection between 'a and 'a option for infinite types\<close>

lemma infinite_option_bijection:
  assumes "infinite (UNIV :: 'a set)" shows "\<exists>f :: 'a \<Rightarrow> 'a option. bij f"
proof -
  from infinite_countable_subset[OF assms] obtain g :: "nat \<Rightarrow> 'a" where g: "inj g" "range g \<subseteq> UNIV" by blast
  let ?f = "\<lambda>x. if x \<in> range g then (case inv g x of 0 \<Rightarrow> None | Suc n \<Rightarrow> Some (g n)) else Some x"
  have "?f x = ?f y \<Longrightarrow> x = y" for x y using g(1) by (auto split: nat.splits if_splits simp: inj_on_def) auto
  moreover have "\<exists>x. ?f x = y" for y
    apply (cases y)
    subgoal using g(1) by (intro exI[of _ "g 0"]) auto
    subgoal for y using g
      apply (cases "y \<in> range g")
      subgoal by (intro exI[of _ "g (Suc (inv g y))"]) auto
      subgoal by auto
    done
  done
  ultimately have "bij ?f" unfolding bij_def surj_def by (intro conjI injI) metis+
  thus ?thesis by blast
qed

definition to_option :: "'a :: infinite \<Rightarrow> 'a option" where
  "to_option = (SOME f. bij f)"

definition from_option :: "'a :: infinite option \<Rightarrow> 'a" where
  "from_option = inv to_option"

lemma bij_from_option: "bij to_option"
  unfolding to_option_def using someI_ex[OF infinite_option_bijection[OF infinite_UNIV]] .

lemma from_to_option_comp[simp]: "from_option \<circ> to_option = id"
  unfolding from_option_def by (intro inv_o_cancel bij_is_inj bij_from_option)

lemma from_to_option[simp]: "from_option (to_option x) = x"
  by (simp add: pointfree_idE)

lemma to_from_option_comp[simp]: "to_option \<circ> from_option = id"
  unfolding from_option_def surj_iff[symmetric] by (intro bij_is_surj bij_from_option)

lemma to_from_option[simp]: "to_option (from_option x) = x"
  by (simp add: pointfree_idE)


lemma var_subst_comp: "t \<cdot> (Var \<circ> f) \<cdot> g = t \<cdot> (g \<circ> f)"
  by (simp add: comp_def subst_subst_compose[symmetric] subst_compose_def del: subst_subst_compose)

subsection \<open>More polymorphic rewriting\<close>

inductive_set rstep' :: "('f, 'v) trs \<Rightarrow> ('f, 'w) term rel" for R where
  rstep' [intro]: "(l, r) \<in> R \<Longrightarrow> (C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep' R"

lemma rstep_eq_rstep': "rstep R = rstep' R"
by (auto elim: rstep'.cases)

lemma rstep'_mono:
  assumes "(s, t) \<in> rstep' R" shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> rstep' R"
proof -
  from assms obtain D l r \<sigma> where "(l, r) \<in> R" "s = D\<langle>l \<cdot> \<sigma>\<rangle>" "t = D\<langle>r \<cdot> \<sigma>\<rangle>" by (auto simp add: rstep'.simps)
  then show ?thesis using rstep'[of l r R "C \<circ>\<^sub>c D" \<sigma>] by simp
qed

lemma ctxt_closed_rstep' [simp]:
  shows "ctxt.closed (rstep' R)"
  by (auto simp: ctxt.closed_def rstep'_mono elim: ctxt.closure.induct)

lemma rstep'_stable:
  assumes "(s, t) \<in> rstep' R" shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> rstep' R"
proof -
  from assms obtain C l r \<tau> where "(l, r) \<in> R" "s = C\<langle>l \<cdot> \<tau>\<rangle>" "t = C\<langle>r \<cdot> \<tau>\<rangle>" by (auto simp add: rstep'.simps)
  then show ?thesis using rstep'[of l r R "C \<cdot>\<^sub>c \<sigma>" "\<tau> \<circ>\<^sub>s \<sigma>"] by simp
qed

lemma rsteps'_stable:
  "(s, t) \<in> (rstep' R)\<^sup>* \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep' R)\<^sup>*"
  by (induct rule: rtrancl_induct) (auto dest: rstep'_stable[of _ _ _ \<sigma>])

lemma rstep'_sub_vars:
  assumes "(s, t) \<in> (rstep' R)\<^sup>*" "wf_trs R"
  shows "vars_term t \<subseteq> vars_term s"
  using assms(1)
proof (induction rule: converse_rtrancl_induct)
  case (step y z)
  obtain l r C \<sigma> where props: "(l, r) \<in> R" "y = C\<langle>l \<cdot> \<sigma>\<rangle>" "z = C\<langle>r \<cdot> \<sigma>\<rangle>" "vars_term r \<subseteq> vars_term l"
    using step(1) assms(2) unfolding wf_trs_def by (auto elim: rstep'.cases) 
  hence "vars_term z \<subseteq> vars_term y"
    using var_cond_stable[OF props(4), of \<sigma>] vars_term_ctxt_apply[of C "_ \<cdot> \<sigma>"] by auto
  thus ?case using step(3) by simp
qed simp

inductive_set rstep_r_p_s' :: "('f, 'v) trs \<Rightarrow> ('f, 'v) rule \<Rightarrow> pos \<Rightarrow> ('f, 'v, 'w) gsubst \<Rightarrow> ('f, 'w) term rel"
  for R r p \<sigma> where
  rstep_r_p_s' [intro]: "r \<in> R \<Longrightarrow> p = hole_pos C \<Longrightarrow> (C\<langle>fst r \<cdot> \<sigma>\<rangle>, C\<langle>snd r \<cdot> \<sigma>\<rangle>) \<in> rstep_r_p_s' R r p \<sigma>"

declare rstep_r_p_s'.cases [elim]

(* TODO: change definition above *)

lemma rstep_r_p_s'I:
  "r \<in> R \<Longrightarrow> p = hole_pos C \<Longrightarrow> s = C\<langle>fst r \<cdot> \<sigma>\<rangle> \<Longrightarrow> t = C\<langle>snd r \<cdot> \<sigma>\<rangle> \<Longrightarrow> (s, t) \<in> rstep_r_p_s' R r p \<sigma>"
  by auto

lemma rstep_r_p_s'E:
  assumes "(s, t) \<in> rstep_r_p_s' R r p \<sigma>"
  obtains C where "r \<in> R" "p = hole_pos C" "s = C\<langle>fst r \<cdot> \<sigma>\<rangle>" "t = C\<langle>snd r \<cdot> \<sigma>\<rangle>"
  using rstep_r_p_s'.cases assms by metis

lemma rstep_r_p_s_eq_rstep_r_p_s': "rstep_r_p_s R r p \<sigma> = rstep_r_p_s' R r p \<sigma>"
  by (auto simp: rstep_r_p_s_def rstep_r_p_s'.simps) (metis hole_pos_ctxt_of_pos_term)

lemma rstep'_iff_rstep_r_p_s':
  "(s, t) \<in> rstep' R \<longleftrightarrow> (\<exists>l r p \<sigma>. (s, t) \<in> rstep_r_p_s' R (l, r) p \<sigma>)"
  by (auto simp: rstep'.simps rstep_r_p_s'.simps)

lemma rstep_r_p_s'_deterministic:
  assumes "wf_trs R" "(s, t) \<in> rstep_r_p_s' R r p \<sigma>" "(s, t') \<in> rstep_r_p_s' R r p \<tau>"
  shows "t = t'"
proof -
  obtain C where 1: "r \<in> R" "s = C\<langle>fst r \<cdot> \<sigma>\<rangle>" "t = C\<langle>snd r \<cdot> \<sigma>\<rangle>" "p = hole_pos C"
    using assms(2) by (auto simp: rstep_r_p_s'.simps)
  obtain D where 2: "s = D\<langle>fst r \<cdot> \<tau>\<rangle>" "t' = D\<langle>snd r \<cdot> \<tau>\<rangle>" "p = hole_pos D"
    using assms(3) by (auto simp: rstep_r_p_s'.simps)
  obtain lhs rhs where 3: "r = (lhs, rhs)" by force
  show ?thesis using 1(1) 2(1,3) unfolding 1(2,3,4) 2(2) 3
  proof (induct C arbitrary: D)
    case Hole have [simp]: "D = \<box>" using Hole by (cases D) auto
    thus ?case using `wf_trs R` Hole by (auto simp: wf_trs_def term_subst_eq_conv)
  next
    case (More f ss1 C' ss2) thus ?case by (cases D) auto
  qed
qed

lemma rstep_r_p_s'_preserves_funas_terms:
  assumes "wf_trs R" "(s, t) \<in> rstep_r_p_s' R r p \<sigma>" "funas_trs R \<subseteq> F" "funas_term s \<subseteq> F"
  shows "funas_term t \<subseteq> F"
proof -
  obtain C where 1: "r \<in> R" "s = C\<langle>fst r \<cdot> \<sigma>\<rangle>" "t = C\<langle>snd r \<cdot> \<sigma>\<rangle>" "p = hole_pos C"
    using assms(2) by (auto simp: rstep_r_p_s'.simps)
  then have "funas_ctxt C \<subseteq> F" using assms(4) by auto
  obtain lhs rhs where 3: "r = (lhs, rhs)" by force
  show ?thesis using 1(1) using assms(1,3,4)
    by (force simp: 1(2,3,4) 3 funas_term_subst funas_trs_def funas_rule_def wf_trs_def)
qed

lemma rstep'_preserves_funas_terms:
  "funas_trs R \<subseteq> F \<Longrightarrow> funas_term s \<subseteq> F \<Longrightarrow> (s, t) \<in> rstep' R \<Longrightarrow> wf_trs R \<Longrightarrow> funas_term t \<subseteq> F"
  unfolding rstep'_iff_rstep_r_p_s' using rstep_r_p_s'_preserves_funas_terms by blast

lemma rstep_r_p_s'_stable:
  "(s, t) \<in> rstep_r_p_s' R r p \<sigma> \<Longrightarrow> (s \<cdot> \<tau>, t \<cdot> \<tau>) \<in> rstep_r_p_s' R r p (\<sigma> \<circ>\<^sub>s \<tau>)"
  by (auto elim!: rstep_r_p_s'E intro!: rstep_r_p_s'I simp: subst_subst simp del: subst_subst_compose)

lemma rstep_r_p_s'_mono:
  "(s, t) \<in> rstep_r_p_s' R r p \<sigma> \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> rstep_r_p_s' R r (hole_pos C <#> p) \<sigma>"
proof (elim rstep_r_p_s'E)
  fix D assume "r \<in> R" "p = hole_pos D" "s = D\<langle>fst r \<cdot> \<sigma>\<rangle>" "t = D\<langle>snd r \<cdot> \<sigma>\<rangle>"
  then show "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> rstep_r_p_s' R r (hole_pos C <#> p) \<sigma>"
    using rstep_r_p_s'I[of r R "hole_pos C <#> p" "C \<circ>\<^sub>c D" "C\<langle>s\<rangle>" \<sigma> "C\<langle>t\<rangle>"] by simp
qed

lemma rstep_r_p_s'_argE:
  assumes "(s, t) \<in> rstep_r_p_s' R (l, r) (i <# p) \<sigma>"
  obtains f ss ti where "s = Fun f ss" "i < length ss" "t = Fun f (ss[i := ti])" "(ss ! i, ti) \<in> rstep_r_p_s' R (l, r) p \<sigma>"
proof -
  thm upd_conv_take_nth_drop
  from assms obtain C where *: "hole_pos C = i <# p" "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>" "(l, r) \<in> R" by (auto elim: rstep_r_p_s'E)
  then obtain f ls D rs where [simp]: "C = More f ls D rs" "i = length ls" by (cases C) auto
  let ?ss = "ls @ D\<langle>l \<cdot> \<sigma>\<rangle> # rs" and ?ts = "ls @ D\<langle>r \<cdot> \<sigma>\<rangle> # rs" and ?ti = "D\<langle>r \<cdot> \<sigma>\<rangle>"
  have "s = Fun f ?ss" "i < length ?ss" "t = Fun f (list_update ?ss i ?ti)" using * by simp_all
  moreover have "(?ss ! i, ?ti) \<in> rstep_r_p_s' R (l, r) p \<sigma>" using * by (auto intro: rstep_r_p_s'I)
  ultimately show ?thesis ..
qed

lemma rstep_r_p_s'_argI:
  assumes "i < length ss" "(ss ! i, ti) \<in> rstep_r_p_s' R r p \<sigma>"
  shows "(Fun f ss, Fun f (ss[i := ti])) \<in> rstep_r_p_s' R r (i <# p) \<sigma>"
  using assms(1) rstep_r_p_s'_mono[OF assms(2), of "More f (take i ss) Hole (drop (Suc i) ss)"]
  by (auto simp: id_take_nth_drop[symmetric] upd_conv_take_nth_drop min_def)

lemma wf_trs_implies_funposs:
  assumes "wf_trs R" "(s, t) \<in> rstep_r_p_s' R (l, r) p \<sigma>"
  shows "p \<in> funposs s"
proof -
  obtain C where *: "(l, r) \<in> R" "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "p = hole_pos C" using assms by auto
  then show ?thesis
    using assms hole_pos_in_filled_funposs[of "l \<cdot> \<sigma>" C Var] by (force simp: wf_trs_def')
qed

lemma NF_Var':
  assumes "wf_trs R"
  shows "(Var x, t) \<notin> rstep' R"
  unfolding rstep'_iff_rstep_r_p_s' by (auto dest: wf_trs_implies_funposs[OF assms])

lemma fill_holes_rsteps:
  assumes "num_holes C = length ss" "num_holes C = length ts"
    "\<And>i. i < length ss \<Longrightarrow> (ss ! i, ts ! i) \<in> (rstep' \<R>)\<^sup>*"
  shows "(fill_holes C ss, fill_holes C ts) \<in> (rstep' \<R>)\<^sup>*"
  using assms
proof (induct C ss ts rule: fill_holes_induct2)
  case (MFun f Cs xs ys)
  show ?case using MFun(1,2,4)
    by (auto intro!: MFun(3) args_steps_imp_steps  simp: partition_by_nth_nth)
qed auto

subsection \<open>@{term partition_by} stuff\<close>

fun partition_by_idcs :: "nat list \<Rightarrow> nat \<Rightarrow> (nat \<times> nat)" where
  "partition_by_idcs (y # ys) i =
    (if i < y then (0, i) else let (j, k) = partition_by_idcs ys (i - y) in (Suc j, k))"

definition partition_by_idx1 (infix "@\<^sub>1" 105) where
  "ys @\<^sub>1 i = fst (partition_by_idcs ys i)"
definition partition_by_idx2 (infix "@\<^sub>2" 105) where
  "ys @\<^sub>2 i = snd (partition_by_idcs ys i)"

lemma partition_by_predicate:
  assumes "i < length xs"
  shows "partition_by (filter P xs) (map (\<lambda>x. if P x then Suc 0 else 0) xs) ! i = (if P (xs ! i) then [xs ! i] else [])"
  using assms by (induct xs arbitrary: i) (auto simp: less_Suc_eq_0_disj)

lemma nth_by_partition_by:
  "length xs = sum_list ys \<Longrightarrow> i < sum_list ys \<Longrightarrow> xs ! i = partition_by xs ys ! ys @\<^sub>1 i ! ys @\<^sub>2 i"
  apply (induct ys arbitrary: i xs)
  apply (auto simp: partition_by_idx1_def partition_by_idx2_def split: prod.splits)
  by (metis add_less_cancel_left diff_add_inverse fst_conv length_drop less_or_eq_imp_le linordered_semidom_class.add_diff_inverse nth_drop snd_conv)

lemma nth_concat_by_shape:
  assumes "ys = map length xss" "i < sum_list ys"
  shows "concat xss ! i = xss ! ys @\<^sub>1 i ! ys @\<^sub>2 i"
  using nth_by_partition_by[of "concat xss" ys i] partition_by_concat_id[of xss ys] assms
  by (auto simp: length_concat)

lemma list_update_by_partition_by:
  "length xs = sum_list ys \<Longrightarrow> i < sum_list ys \<Longrightarrow>
   xs[i := x] = concat (partition_by xs ys[ys @\<^sub>1 i := (partition_by xs ys ! ys @\<^sub>1 i)[ys @\<^sub>2 i := x]])"
proof (induct ys arbitrary: i xs)
  case (Cons y ys) show ?case
    using list_update_append[of "take y xs" "drop y xs" i x] Cons(1)[of "drop y xs" "i - y"] Cons(2,3)
    by (auto simp: partition_by_idx1_def partition_by_idx2_def less_diff_conv2 min.absorb2 split: prod.splits)
qed auto

lemma list_update_concat_by_shape:
  assumes "ys = map length xss" "i < sum_list ys"
  shows "concat xss[i := x] = concat (xss[ys @\<^sub>1 i := (xss ! ys @\<^sub>1 i)[ys @\<^sub>2 i := x]])"
  using list_update_by_partition_by[of "concat xss" ys i] partition_by_concat_id[of xss ys] assms
  by (auto simp: length_concat)

lemma partition_by_idx1_bound:
  "i < sum_list ys \<Longrightarrow> ys @\<^sub>1 i < length ys"
  apply (induct ys arbitrary: i)
  apply (auto simp: partition_by_idx1_def split: prod.splits)
  by (metis add_diff_inverse_nat add_less_imp_less_left fst_conv)

lemma partition_by_idx2_bound:
  "i < sum_list ys \<Longrightarrow> ys @\<^sub>2 i <  ys ! ys @\<^sub>1 i"
  apply (induct ys arbitrary: i)
  apply (auto simp: partition_by_idx1_def partition_by_idx2_def split: prod.splits)
  by (metis add_diff_inverse_nat fst_conv nat_add_left_cancel_less snd_conv)

subsection \<open>All positions of a multi-hole context\<close>

fun all_poss_mctxt :: "('f, 'v) mctxt \<Rightarrow> pos set"
where
  "all_poss_mctxt (MVar x) = {\<epsilon>}"
| "all_poss_mctxt MHole = {\<epsilon>}"
| "all_poss_mctxt (MFun f cs) = {\<epsilon>} \<union> \<Union>(set (map (\<lambda> i. (\<lambda> p. i <# p) ` all_poss_mctxt (cs ! i)) [0 ..< length cs]))"

lemma all_poss_mctxt_simp [simp]:
  "all_poss_mctxt (MFun f cs) = {\<epsilon>} \<union> {i <# p | i p. i < length cs \<and> p \<in> all_poss_mctxt (cs ! i)}"
  by auto

declare all_poss_mctxt.simps(3)[simp del]

lemma all_poss_mctxt_conv:
  "all_poss_mctxt C = poss_mctxt C \<union> hole_poss C"
  by (induct C) auto

lemma root_in_all_poss_mctxt[simp]:
  "\<epsilon> \<in> all_poss_mctxt C"
  by (cases C) auto

lemma hole_poss_mctxt_of_term[simp]:
  "hole_poss (mctxt_of_term t) = {}"
  by (induct t) auto

lemma poss_mctxt_mctxt_of_term[simp]:
  "poss_mctxt (mctxt_of_term t) = poss t"
  by (induct t) auto

lemma all_poss_mctxt_mctxt_of_term[simp]:
  "all_poss_mctxt (mctxt_of_term t) = poss t"
  by (induct t) auto

lemma mctxt_of_term_leq_imp_eq:
  "mctxt_of_term t \<le> C \<longleftrightarrow> mctxt_of_term t = C"
  by (induct t arbitrary: C) (auto elim!: less_eq_mctxtE1 simp: map_nth_eq_conv)

lemma mctxt_of_term_inj:
  "mctxt_of_term s = mctxt_of_term t \<longleftrightarrow> s = t"
  by (induct s arbitrary: t; case_tac t) (auto simp: map_eq_conv' intro: nth_equalityI)

lemma all_poss_mctxt_map_vars_mctxt [simp]:
  "all_poss_mctxt (map_vars_mctxt f C) = all_poss_mctxt C"
  by (induct C) auto

lemma fill_holes_mctxt_extends_all_poss:
  assumes "length Ds = num_holes C" shows "all_poss_mctxt C \<subseteq> all_poss_mctxt (fill_holes_mctxt C Ds)"
  using assms[symmetric] by (induct C Ds rule: fill_holes_induct) force+

subsection \<open>More operations on multi-hole contexts\<close>

fun root_mctxt :: "('f, 'v) mctxt \<Rightarrow> ('f \<times> nat) option" where
   "root_mctxt MHole = None"
 | "root_mctxt (MVar x) = None"
 | "root_mctxt (MFun f Cs) = Some (f, length Cs)"

fun mreplace_at :: "('f, 'v) mctxt \<Rightarrow> pos \<Rightarrow> ('f, 'v) mctxt \<Rightarrow> ('f, 'v) mctxt" where
  "mreplace_at C \<epsilon> D = D"
| "mreplace_at (MFun f Cs) (i <# p) D = MFun f (take i Cs @ mreplace_at (Cs ! i) p D # drop (i+1) Cs)"
(* Should use @{term "Cs[i := mreplace_at (Cs ! i)"}? see also @{thm upd_conv_take_nth_drop} *)

fun subm_at :: "('f, 'v) mctxt \<Rightarrow> pos \<Rightarrow> ('f, 'v) mctxt" where
  "subm_at C \<epsilon> = C"
| "subm_at (MFun f Cs) (i <# p) = subm_at (Cs ! i) p"

lemma subm_at_hole_poss[simp]:
  "p \<in> hole_poss C \<Longrightarrow> subm_at C p = MHole"
  by (induct C arbitrary: p) auto

lemma subm_at_mctxt_of_term:
  "p \<in> poss t \<Longrightarrow> subm_at (mctxt_of_term t) p = mctxt_of_term (subt_at t p)"
  by (induct t arbitrary: p) auto

lemma subm_at_mreplace_at[simp]:
  "p \<in> all_poss_mctxt C \<Longrightarrow> subm_at (mreplace_at C p D) p = D"
  by (induct C arbitrary: p) (auto simp: nth_append_take)

lemma replace_at_subm_at[simp]:
  "p \<in> all_poss_mctxt C \<Longrightarrow> mreplace_at C p (subm_at C p) = C"
  by (induct C arbitrary: p) (auto simp: id_take_nth_drop[symmetric])

lemma all_poss_mctxt_mreplace_atI1:
  "p \<in> all_poss_mctxt C \<Longrightarrow> q \<in> all_poss_mctxt C \<Longrightarrow> \<not>(q > p) \<Longrightarrow> q \<in> all_poss_mctxt (mreplace_at C p D)"
proof (induct C arbitrary: p q)
  let ?hd = "\<lambda>p. (case p :: pos of i <# _ \<Rightarrow> i)"
  case (MFun f Cs) then show ?case
    by (cases "?hd p = ?hd q") (auto simp: nth_append_take less_pos_def nth_append_drop_is_nth_conv nth_append_take_drop_is_nth_conv)
qed auto

lemma funas_mctxt_sup_mctxt:
  "(C, D) \<in> comp_mctxt \<Longrightarrow> funas_mctxt (C \<squnion> D) = funas_mctxt C \<union> funas_mctxt D"
  by (induct C D rule: comp_mctxt.induct) (auto simp: zip_nth_conv Un_Union_image)

lemma mctxt_of_term_not_hole [simp]:
  "mctxt_of_term t \<noteq> MHole"
  by (cases t) auto

lemma funas_mctxt_mctxt_of_term [simp]:
  "funas_mctxt (mctxt_of_term t) = funas_term t"
  by (induct t) auto

lemma funas_mctxt_mreplace_at:
  assumes "p \<in> all_poss_mctxt C"
  shows "funas_mctxt (mreplace_at C p D) \<subseteq> funas_mctxt C \<union> funas_mctxt D"
  using assms
proof (induct C p D rule: mreplace_at.induct)
  case (2 f Cs i p D) then have Cs: "Cs = take i Cs @ Cs ! i # drop (Suc i) Cs"
    by (auto simp: id_take_nth_drop)
  show ?case using 2 by (subst (2) Cs) auto
qed auto

lemma funas_mctxt_mreplace_at_hole:
  assumes "p \<in> hole_poss C"
  shows "funas_mctxt (mreplace_at C p D) = funas_mctxt C \<union> funas_mctxt D" (is "?L = ?R")
proof
  show "?R \<subseteq> ?L" using assms
  proof (induct C p D rule: mreplace_at.induct)
    case (1 C D) thus ?case by (cases C) auto
  next
    case (2 f Cs i p D) then have Cs: "Cs = take i Cs @ Cs ! i # drop (Suc i) Cs"
      by (auto simp: id_take_nth_drop)
    show ?case using 2 by (subst (1) Cs) auto
  qed auto
next
  show "?L \<subseteq> ?R" using assms by (auto simp: all_poss_mctxt_conv funas_mctxt_mreplace_at)
qed

fun mctxt_term_conv :: "('f, 'v) mctxt \<Rightarrow> ('f, 'v option) term" where
  "mctxt_term_conv MHole = Var None"
| "mctxt_term_conv (MVar v) = Var (Some v)"
| "mctxt_term_conv (MFun f Cs) = Fun f (map mctxt_term_conv Cs)"

fun term_mctxt_conv :: "('f, 'v option) term \<Rightarrow> ('f, 'v) mctxt" where
  "term_mctxt_conv (Var None) = MHole"
| "term_mctxt_conv (Var (Some v)) = MVar v"
| "term_mctxt_conv (Fun f ts) = MFun f (map term_mctxt_conv ts)"

lemma mctxt_term_conv_inv [simp]:
  "mctxt_term_conv (term_mctxt_conv t) = t"
  by (induct t rule: term_mctxt_conv.induct) (auto simp: map_idI)

lemma term_mctxt_conv_inv [simp]:
  "term_mctxt_conv (mctxt_term_conv t) = t"
  by (induct t rule: mctxt_term_conv.induct) (auto simp: map_idI)
  
lemma mctxt_term_conv_bij:
  "bij mctxt_term_conv"
  by (auto intro!: o_bij[of term_mctxt_conv mctxt_term_conv])

lemma term_mctxt_conv_bij:
  "bij term_mctxt_conv"
  by (auto intro!: o_bij[of mctxt_term_conv term_mctxt_conv])

lemma mctxt_term_conv_mctxt_of_term[simp]:
  "mctxt_term_conv (mctxt_of_term t) = t \<cdot> (Var \<circ> Some)"
  by (induct t) auto

lemma term_mctxt_conv_mctxt_of_term_conv:
  "term_mctxt_conv (t \<cdot> (Var \<circ> Some)) = mctxt_of_term t"
  by (induct t) auto

lemma weak_match_mctxt_term_conv_mono:
  "C \<le> D \<Longrightarrow> weak_match (mctxt_term_conv D) (mctxt_term_conv C)"
  by (induct C D rule: less_eq_mctxt_induct) auto

definition term_of_mctxt_subst where "term_of_mctxt_subst = case_option (term_of_mctxt MHole) Var"

lemma term_of_mctxt_to_mctxt_term_conv:
  "term_of_mctxt C = mctxt_term_conv C \<cdot> term_of_mctxt_subst"
  by (induct C) (auto simp: term_of_mctxt_subst_def)

lemma poss_mctxt_term_conv[simp]:
  "poss (mctxt_term_conv C) = all_poss_mctxt C"
  by (induct C) auto

lemma funas_term_mctxt_term_conv[simp]:
  "funas_term (mctxt_term_conv C) = funas_mctxt C"
  by (induct C) auto

lemma all_poss_mctxt_term_mctxt_conv[simp]:
  "all_poss_mctxt (term_mctxt_conv t) = poss t"
  by (induct t rule: term_mctxt_conv.induct) auto

lemma funas_mctxt_term_mctxt_conv[simp]:
  "funas_mctxt (term_mctxt_conv t) = funas_term t"
  by (induct t rule: term_mctxt_conv.induct) auto

lemma subm_at_term_mctxt_conv:
  "p \<in> poss t \<Longrightarrow> subm_at (term_mctxt_conv t) p = term_mctxt_conv (subt_at t p)"
  by (induct t p rule: subt_at.induct) auto

lemma subt_at_mctxt_term_conv:
  "p \<in> all_poss_mctxt C \<Longrightarrow> subt_at (mctxt_term_conv C) p = mctxt_term_conv (subm_at C p)"
  by (induct C p rule: subm_at.induct) auto

lemma subm_at_subt_at_conv:
  "p \<in> all_poss_mctxt C \<Longrightarrow> subm_at C p = term_mctxt_conv (subt_at (mctxt_term_conv C) p)"
  by (induct C p rule: subm_at.induct) auto

lemma mctxt_term_conv_fill_holes_mctxt:
  assumes "num_holes C = length Cs"
  shows "mctxt_term_conv (fill_holes_mctxt C Cs) = fill_holes (map_vars_mctxt Some C) (map mctxt_term_conv Cs)"
  using assms by (induct C Cs rule: fill_holes_induct) (auto simp: comp_def)

lemma map_vars_mctxt_fill_holes_mctxt:
  assumes "num_holes C = length Cs"
  shows "map_vars_mctxt f (fill_holes_mctxt C Cs) = fill_holes_mctxt (map_vars_mctxt f C) (map (map_vars_mctxt f) Cs)"
  using assms by (induct C Cs rule: fill_holes_induct) (auto simp: comp_def)

lemma mctxt_term_conv_map_vars_mctxt_subst:
  shows "mctxt_term_conv (map_vars_mctxt f C) = mctxt_term_conv C \<cdot> (Var \<circ> map_option f)"
  by (induct C) auto

lemma map_vars_mctxt_map_vars_mctxt[simp]:
  shows "map_vars_mctxt f (map_vars_mctxt g C) = map_vars_mctxt (f \<circ> g) C"
  by (induct C) auto

lemma funas_mctxt_fill_holes_mctxt:
  assumes "num_holes C = length ts"
  shows "funas_term (fill_holes C ts) = funas_mctxt C \<union> \<Union>(set (map funas_term ts))"
  using funas_term_fill_holes_iff[OF assms] by auto

(* an mctxt version of fill_holes_mctxt_fill_holes *)
lemma fill_fill_holes_mctxt:
  assumes "length Cs' = num_holes L'" "length Cs = num_holes (fill_holes_mctxt L' Cs')"
  shows "fill_holes_mctxt (fill_holes_mctxt L' Cs') Cs = fill_holes_mctxt L'
     (map (\<lambda>(D, Es). fill_holes_mctxt D Es) (zip Cs' (partition_holes Cs Cs')))" (is "?L = ?R")
proof -
  note fill_holes_mctxt_fill_holes
  have "fill_holes (fill_holes_mctxt (map_vars_mctxt Some L') (map (map_vars_mctxt Some) Cs'))
    (map mctxt_term_conv Cs) = fill_holes (map_vars_mctxt Some L')
    (map (\<lambda>x. mctxt_term_conv (fill_holes_mctxt (Cs' ! x) (partition_holes Cs Cs' ! x))) [0..<num_holes L'])"
    using assms by (subst fill_holes_mctxt_fill_holes)
      (auto simp: comp_def mctxt_term_conv_fill_holes_mctxt length_partition_by_nth num_holes_fill_holes_mctxt intro!: arg_cong[of _ _ "fill_holes _"])
  then have "term_mctxt_conv (mctxt_term_conv ?L) = term_mctxt_conv (mctxt_term_conv ?R)"
    using assms by (intro arg_cong[of _ _ "term_mctxt_conv"]) (auto simp: zip_nth_conv comp_def
      mctxt_term_conv_fill_holes_mctxt map_vars_mctxt_fill_holes_mctxt)
  then show ?thesis by simp
qed

inductive_set mrstep :: "('f, 'w) trs \<Rightarrow> ('f, 'v) mctxt rel" for R where
  mrstep [intro]: "(mctxt_term_conv C, mctxt_term_conv D) \<in> rstep' R \<Longrightarrow> (C, D) \<in> mrstep R"

lemma mrstepD: "(C, D) \<in> mrstep R \<Longrightarrow> (mctxt_term_conv C, mctxt_term_conv D) \<in> rstep' R"
  by (rule mrstep.induct)

lemma mrstepI_inf: (* caveat: note the implied 'v :: infinite *)
  assumes "(mctxt_term_conv C \<cdot> (Var \<circ> from_option), mctxt_term_conv D \<cdot> (Var \<circ> from_option)) \<in> rstep R"
  shows "(C, D) \<in> mrstep R"
  using rstep'_stable[OF assms[unfolded rstep_eq_rstep'], of "Var \<circ> to_option"]
  by (intro mrstep.intros) (auto simp del: subst_subst_compose simp: subst_subst subst_compose_def)

lemma mrstepD_inf: (* caveat: note the implied 'v :: infinite *)
  "(C, D) \<in> mrstep R \<Longrightarrow> (mctxt_term_conv C \<cdot> (Var \<circ> from_option), mctxt_term_conv D \<cdot> (Var \<circ> from_option)) \<in> rstep R"
  by (metis rstep'_stable mrstepD rstep_eq_rstep')

lemma NF_MVar_MHole: 
  assumes "wf_trs R" and "C = MVar x \<or> C = MHole" 
  shows "(C, D) \<notin> mrstep R"
  using NF_Var'[OF assms(1)] assms(2) by (force dest: mrstepD)

lemma mctxt_neq_mholeE:
  "x \<noteq> MHole \<Longrightarrow> (\<And>v. x = MVar v \<Longrightarrow> P) \<Longrightarrow> (\<And>f Cs. x = MFun f Cs \<Longrightarrow> P) \<Longrightarrow> P"
  by (cases x) auto

lemma leq_mholeE:
  "C \<le> MHole \<Longrightarrow> (C = MHole \<Longrightarrow> P) \<Longrightarrow> P"
  by (cases C) (auto simp: less_eq_mctxt_def)

lemma leq_mvarE:
  "C \<le> MVar v \<Longrightarrow> (C = MHole \<Longrightarrow> P) \<Longrightarrow> (C = MVar v \<Longrightarrow> P) \<Longrightarrow> P"
  by (cases C) (auto simp: less_eq_mctxt_def split_ifs)

lemma leq_mfunE:
  "C \<le> MFun f Ds \<Longrightarrow> (C = MHole \<Longrightarrow> P) \<Longrightarrow>
  (\<And>Cs. C = MFun f Cs \<Longrightarrow> length Cs = length Ds \<Longrightarrow> (\<And>i. i < length Ds \<Longrightarrow> Cs ! i \<le> Ds ! i) \<Longrightarrow> P) \<Longrightarrow> P"
  by (cases C) (auto simp: less_eq_mctxt_def split_ifs list_eq_iff_nth_eq)

lemma prefix_comp_mctxt:
  "C \<le> E \<Longrightarrow> D \<le> E \<Longrightarrow> (C, D) \<in> comp_mctxt"
proof (induct E arbitrary: C D)
  case (MFun f Es C D)
  then show ?case
  proof (elim leq_mfunE)
    fix Cs Ds
    assume C: "C = MFun f Cs" and D: "D = MFun f Ds"
      and lC: "length Cs = length Es" and lD: "length Ds = length Es"
      and Ci: "\<And>i. i < length Es \<Longrightarrow> Cs ! i \<le> Es ! i" and Di: "\<And>i. i < length Es \<Longrightarrow> Ds ! i \<le> Es ! i"
      and IH: "\<And>E' C' D'. E' \<in> set Es \<Longrightarrow> C' \<le> E' \<Longrightarrow> D' \<le> E' \<Longrightarrow> (C', D') \<in> comp_mctxt"
    show "(C, D) \<in> comp_mctxt"
    by (auto simp: C D lC lD intro!: comp_mctxt.intros IH[OF _ Ci Di])
  qed (auto intro: comp_mctxt.intros)
qed (auto elim: leq_mholeE leq_mvarE intro: comp_mctxt.intros)

lemma less_eq_mctxt_sup_conv1:
  "(C, D) \<in> comp_mctxt \<Longrightarrow> C \<le> D \<longleftrightarrow> C \<squnion> D = D"
  by (induct C D rule: comp_mctxt.induct) (auto elim!: less_eq_mctxtE2 nth_equalityE intro: nth_equalityI less_eq_mctxtI2(3))

lemma less_eq_mctxt_sup_conv2:
  "(C, D) \<in> comp_mctxt \<Longrightarrow> D \<le> C \<longleftrightarrow> C \<squnion> D = C"
  using less_eq_mctxt_sup_conv1[OF comp_mctxt_sym] by (auto simp: ac_simps)

lemma comp_mctxt_mctxt_of_term1[dest!]:
  "(C, mctxt_of_term t) \<in> comp_mctxt \<Longrightarrow> C \<le> mctxt_of_term t"
  by (induct C "mctxt_of_term t" arbitrary: t rule: comp_mctxt.induct; case_tac t) (auto intro: less_eq_mctxtI2)

lemmas comp_mctxt_mctxt_of_term2[dest!] = comp_mctxt_mctxt_of_term1[OF comp_mctxt_sym]

lemma mfun_leq_mfunI:
  "f = g \<Longrightarrow> length Cs = length Ds \<Longrightarrow> (\<And>i. i < length Ds \<Longrightarrow> Cs ! i \<le> Ds ! i) \<Longrightarrow> MFun f Cs \<le> MFun g Ds"
  by (auto simp: less_eq_mctxt_def list_eq_iff_nth_eq)

lemma prefix_mctxt_sup:
  assumes "C \<le> (E :: ('f, 'v) mctxt)" "D \<le> E" shows "C \<squnion> D \<le> E"
  using assms
  by (induct E arbitrary: C D) (auto elim!: leq_mholeE leq_mvarE leq_mfunE intro!: mfun_leq_mfunI)

lemma mreplace_at_leqI:
  "p \<in> all_poss_mctxt C \<Longrightarrow> C \<le> E \<Longrightarrow> D \<le> subm_at E p \<Longrightarrow> mreplace_at C p D \<le> E"
  by (induct C p D arbitrary: E rule: mreplace_at.induct)
    (auto elim!: less_eq_mctxtE1 intro!: less_eq_mctxtI1 simp: upd_conv_take_nth_drop[symmetric] nth_list_update)

definition funposs_mctxt :: "('f, 'v) mctxt \<Rightarrow> pos set" where
  "funposs_mctxt C = funposs (mctxt_term_conv C)"

lemma funposs_mctxt_subset_poss_mctxt:
  "funposs_mctxt C \<subseteq> poss_mctxt C"
  by (induct C) (force simp: funposs_mctxt_def)+

lemma funposs_mctxt_subset_all_poss_mctxt:
  "funposs_mctxt C \<subseteq> all_poss_mctxt C"
  by (induct C) (force simp: funposs_mctxt_def)+

lemma funposs_mctxt_mono:
  "C \<le> D \<Longrightarrow> p \<in> funposs_mctxt C \<Longrightarrow> p \<in> funposs_mctxt D"
  unfolding less_eq_mctxt_prime funposs_mctxt_def
  by (induct C D arbitrary: p rule: less_eq_mctxt'.induct) auto

lemma funposs_mctxt_compat:
  "C \<le> D \<Longrightarrow> p \<in> poss_mctxt C \<Longrightarrow> p \<in> funposs_mctxt D \<Longrightarrow> p \<in> funposs_mctxt C"
  unfolding less_eq_mctxt_prime funposs_mctxt_def
  by (induct C D arbitrary: p rule: less_eq_mctxt'.induct) auto

lemma funposs_mctxt_mctxt_of_term[simp]:
  "funposs_mctxt (mctxt_of_term t) = funposs t"
  by (induct t) (auto simp: funposs_mctxt_def)

lemma prefix_and_fewer_holes_implies_equal_mctxt:
  "C \<le> D \<Longrightarrow> hole_poss C \<subseteq> hole_poss D \<Longrightarrow> C = D"
proof (induct C D rule: less_eq_mctxt_induct)
  case (1 D) then show ?case by (cases D) auto
next
  case (3 Cs Ds f)
  have "i < length Ds \<Longrightarrow> hole_poss (Cs ! i) \<subseteq> hole_poss (Ds ! i)" for i using 3(1,4) by auto
  then show ?case using 3 by (auto intro!: nth_equalityI)
qed auto

lemma compare_mreplace_at:
  "p \<in> poss_mctxt C \<Longrightarrow> mreplace_at C p D \<le> mreplace_at C p E \<longleftrightarrow> D \<le> E"
  apply (induct C arbitrary: p; case_tac p)
  apply (auto elim!: less_eq_mctxtE2(3) intro!: less_eq_mctxt_intros(3) simp: nth_append nth_Cons'
    split: if_splits)
  by (metis Nat.add_diff_assoc Suc_leI less_iff_Suc_add less_imp_le_nat less_irrefl_nat min.absorb2 order_refl)

lemma proper_prefix_hole_poss_imp_funposs:
  assumes "p \<in> hole_poss C" "q < p"
  shows "q \<in> funposs_mctxt C"
  using assms
  apply (induct C arbitrary: p q; case_tac p; case_tac q)
  apply (auto simp: funposs_mctxt_def)
  apply (metis (no_types, hide_lams) lessThan_iff less_simps(4) nth_mem pos.exhaust)+
  done

lemma merge_mreplace_at:
  "p \<in> poss_mctxt C \<Longrightarrow> mreplace_at C p (D \<squnion> E) =  mreplace_at C p D \<squnion> mreplace_at C p E"
  by (induct C arbitrary: p; case_tac p) (auto intro: nth_equalityI)

lemma compare_mreplace_atI:
  "C \<le> D \<Longrightarrow> C' \<le> D' \<Longrightarrow> p \<in> poss_mctxt C \<Longrightarrow> mreplace_at C p C' \<le> mreplace_at D p D'"
  by (induct C D arbitrary: p rule: less_eq_mctxt_induct; case_tac p)
    (auto intro!: less_eq_mctxt_intros(3) simp: nth_append nth_Cons')

(* TODO: consolidate with compare_mreplace_atI *)
lemma compare_mreplace_atI':
  "C \<le> D \<Longrightarrow> C' \<le> D' \<Longrightarrow> p \<in> all_poss_mctxt C \<Longrightarrow> mreplace_at C p C' \<le> mreplace_at D p D'"
  by (induct C D arbitrary: p rule: less_eq_mctxt_induct; case_tac p)
    (auto intro!: less_eq_mctxt_intros(3) simp: nth_append nth_Cons')

subsection \<open>finiteness of prefixes\<close>

lemma finite_set_Cons:
  assumes A: "finite A" and B: "finite B"
  shows "finite (set_Cons A B)"
proof -
  have "set_Cons A B = case_prod (op #) ` (A \<times> B)" by (auto simp: set_Cons_def)
  then show ?thesis
    by (simp add: finite_imageI[OF finite_cartesian_product[OF A B],of "case_prod (op #)"])
qed

lemma listset_finite:
  assumes "\<forall>A \<in> set As. finite A"
  shows "finite (listset As)"
  using assms
  by (induct As) (auto simp: finite_set_Cons)

lemma elem_listset:
  "xs \<in> listset As = (length xs = length As \<and> (\<forall>i < length As. xs ! i \<in> As ! i))"
proof (induct As arbitrary: xs)
  case (Cons A As xs) thus ?case
    by (cases xs) (auto simp: set_Cons_def nth_Cons nat.splits)
qed auto

lemma finite_pre_mctxt:
  fixes C :: "('f, 'v) mctxt"
  shows "finite { N. N \<le> C }"
proof (induct C)
  case MHole
  have *: "{ N. N \<le> MHole } = { MHole }" by (auto simp: less_eq_mctxt_def)
  show ?case by (simp add: *)
next
  case (MVar x)
  have *: "{ N. N \<le> MVar x } = { MHole, MVar x }"
    by (auto simp: less_eq_mctxt_def split_ifs elim: mctxt_neq_mholeE)
  show ?case by (simp add: *)
next
  case (MFun f Cs)
  have *: "{ N. N \<le> MFun f Cs } = { MHole } \<union> MFun f ` { Ds. Ds \<in> listset (map (\<lambda>C. { N. N \<le> C }) Cs)}"
  unfolding elem_listset
    by (auto simp: image_def less_eq_mctxt_def split_ifs list_eq_iff_nth_eq elim!: mctxt_neq_mholeE) auto
  show ?case
    by (auto simp: * MFun listset_finite[of "map (\<lambda>C. { N. N \<le> C }) Cs"])
qed

text \<open>well-founded-ness of <\<close>

fun mctxt_syms :: "('f, 'v) mctxt \<Rightarrow> nat" where
  "mctxt_syms MHole = 0"
| "mctxt_syms (MVar v) = 1"
| "mctxt_syms (MFun f Cs) = 1 + sum_list (map mctxt_syms Cs)" 
  
lemma mctxt_syms_mono:
  "C \<le> D \<Longrightarrow> mctxt_syms C \<le> mctxt_syms D"
by (induct D arbitrary: C; elim leq_mholeE leq_mvarE leq_mfunE)
  (auto simp: map_upt_len_conv[of mctxt_syms,symmetric] intro: sum_list_mono)

lemma sum_list_strict_mono_aux:
  fixes xs ys :: "nat list"
  shows "length xs = length ys \<Longrightarrow> (\<And>i. i < length ys \<Longrightarrow> xs ! i \<le> ys ! i) \<Longrightarrow> sum_list xs < sum_list ys \<or> xs = ys"
proof (induct xs arbitrary: ys)
  case (Cons x xs zs) note * = this
  then show ?case
  proof (cases zs)
    case (Cons y ys)
    have hd: "x \<le> y" and tl: "\<And>i. i < length ys \<Longrightarrow> xs ! i \<le> ys ! i"
      using *(3) by (auto simp: Cons nth_Cons nat.splits)
    show ?thesis using hd *(1)[OF _ tl] *(2) by (auto simp: nth_Cons Cons)
  qed auto
qed auto

lemma mctxt_syms_strict_mono[simp]:
  "C < D \<Longrightarrow> mctxt_syms C < mctxt_syms D"
proof -
  assume "C < D"
  also have "C \<le> D \<Longrightarrow> C = D \<or> mctxt_syms C < mctxt_syms D"
  proof ((induct D arbitrary: C; elim leq_mholeE leq_mvarE leq_mfunE), goal_cases)
    case (5 f Ds C Cs)
    have "i < length Ds \<Longrightarrow> mctxt_syms (Cs ! i) \<le> mctxt_syms (Ds ! i)" for i
        using 5(1)[OF _ 5(4), of i] by auto
    then show "C = MFun f Ds \<or> mctxt_syms C < mctxt_syms (MFun f Ds)"
    using 5(1)[OF _ 5(4)] sum_list_strict_mono_aux[of "map mctxt_syms Cs" "map mctxt_syms Ds"]
      by (auto simp: list_eq_iff_nth_eq 5(2,3))
  qed auto
  ultimately show ?thesis by (auto simp: less_mctxt_def)
qed

lemma wf_less_mctxt [simp]:
  "wf { (C :: ('f, 'v) mctxt, D). C < D }"
  by (rule wf_subset[of "inv_image { (a, b). a < b } mctxt_syms"]) (auto simp: wf_less)

lemma map_zip2 [simp]:
  "n = length xs \<Longrightarrow> zip xs (replicate n y) = map (\<lambda>x. (x,y)) xs"
  by (induct xs arbitrary: n) auto

lemma map_zip1 [simp]:
  "n = length ys \<Longrightarrow> zip (replicate n x) ys = map (\<lambda>y. (x,y)) ys"
  by (induct ys arbitrary: n) auto

lemma all_poss_mctxt_mono:
  "C \<le> D \<Longrightarrow> all_poss_mctxt C \<subseteq> all_poss_mctxt D"
  by (induct C D rule: less_eq_mctxt_induct) force+

lemma all_poss_mctxt_inf_mctxt:
  "(C, D) \<in> comp_mctxt \<Longrightarrow> all_poss_mctxt (C \<sqinter> D) = all_poss_mctxt C \<inter> all_poss_mctxt D"
  by (induct C D rule: comp_mctxt.induct) auto

lemma less_eq_subm_at:
  "p \<in> all_poss_mctxt C \<Longrightarrow> C \<le> D \<Longrightarrow> subm_at C p \<le> subm_at D p"
  by (induct C arbitrary: p D) (auto elim: less_eq_mctxtE1)

lemma inf_subm_at:
  "p \<in> all_poss_mctxt (C \<sqinter> D) \<Longrightarrow> subm_at (C \<sqinter> D) p = subm_at C p \<sqinter> subm_at D p"
proof (induct C D arbitrary: p rule: inf_mctxt.induct)
  case (4 f Cs g Ds p) show ?case using 4(1) 4(2)
    by (auto 4 4 intro!: 4(1)[of "(Cs ! i, Ds ! i)" "Cs ! i" "Ds ! i" for i] simp: set_zip)
qed auto

lemma less_eq_fill_holesI:
  assumes "length Ds = num_holes C" "length Es = num_holes C"
    "\<And>i. i < num_holes C \<Longrightarrow> Ds ! i \<le> Es ! i"
  shows "fill_holes_mctxt C Ds \<le> fill_holes_mctxt C Es"
  using assms(1,2)[symmetric] assms(3)
  by (induct C Ds Es rule: fill_holes_induct2) (auto intro!: less_eq_mctxtI1 simp: partition_by_nth_nth)

lemma less_eq_fill_holesD:
  assumes "length Ds = num_holes C" "length Es = num_holes C"
    "fill_holes_mctxt C Ds \<le> fill_holes_mctxt C Es" "i < num_holes C"
  shows "Ds ! i \<le> Es ! i"
  using assms(1,2)[symmetric] assms(3,4)
proof (induct C Ds Es arbitrary: i rule: fill_holes_induct2)
  case (MFun f Cs Ds Es)
  obtain j k where "j < length Cs" "k < num_holes (Cs ! j)"
    "zip Ds Es ! i = partition_holes (zip Ds Es) Cs ! j ! k"
    using nth_concat_split[of i "partition_holes (zip Ds Es) Cs"] MFun(1,2,5) by auto
  moreover then have "f (zip Ds Es ! i) = partition_holes (map f (zip Ds Es)) Cs ! j ! k" for f
    using nth_map[of k "partition_holes (zip Ds Es) Cs ! j" f] MFun(1,2)
      length_partition_by_nth[of "map num_holes Cs" "zip Ds Es"] by simp
  from this[of fst] this[of snd] map_fst_zip[of Ds Es] map_snd_zip[of Ds Es]
    have "Ds ! i = partition_holes Ds Cs ! j ! k" "Es ! i = partition_holes Es Cs ! j ! k"
    using MFun(1,2,5) by simp_all
  ultimately show ?case using MFun(3)[of j k] MFun(1,2,4) by (auto elim: less_eq_mctxtE1)
qed auto

lemma less_eq_fill_holes_iff:
  assumes "length Ds = num_holes C" "length Es = num_holes C"
  shows "fill_holes_mctxt C Ds \<le> fill_holes_mctxt C Es \<longleftrightarrow> (\<forall>i < num_holes C. Ds ! i \<le> Es ! i)"
  using assms by (auto intro: less_eq_fill_holesI dest: less_eq_fill_holesD)

lemma fill_holes_mctxt_suffix[simp]:
  assumes "length Ds = num_holes C" shows "C \<le> fill_holes_mctxt C Ds"
  using assms(1)[symmetric]
  by (induct C Ds rule: fill_holes_induct) (auto simp: less_eq_mctxt_def intro: nth_equalityI)

lemma fill_holes_mctxt_id:
  assumes "length Ds = num_holes C" "C = fill_holes_mctxt C Ds" shows "set Ds \<subseteq> {MHole}"
  using assms(1)[symmetric] assms(2)
  apply (induct C Ds rule: fill_holes_induct)
  unfolding set_concat
  by (auto simp: set_conv_nth[of "partition_holes _ _"] list_eq_iff_nth_eq[of _ "map _ _"])

lemma fill_holes_suffix[simp]:
  "num_holes C = length ts \<Longrightarrow> C \<le> mctxt_of_term (fill_holes C ts)"
  by (induct C ts rule: fill_holes_induct) (auto intro: less_eq_mctxtI1)

subsection \<open>Hole positions, left to right\<close>

fun hole_poss' :: "('f, 'v) mctxt \<Rightarrow> pos list" where
  "hole_poss' (MVar x) = []"
| "hole_poss' MHole = [\<epsilon>]"
| "hole_poss' (MFun f cs) = concat (map (\<lambda>i . map (op <# i) (hole_poss' (cs ! i))) [0..<length cs])"

lemma set_hole_poss': "set (hole_poss' C) = hole_poss C"
  by (induct C) auto

lemma length_hole_poss'[simp]: "length (hole_poss' C) = num_holes C"
  by (induct C) (auto simp: length_concat intro!: arg_cong[of _ _ sum_list] nth_equalityI)

lemma hole_poss'_map_vars_mctxt[simp]:
 "hole_poss' (map_vars_mctxt f C) = hole_poss' C"
  by (induct C rule: hole_poss'.induct) (auto intro: arg_cong[of _ _ concat])

lemma subt_at_fill_holes:
  assumes "length ts = num_holes C" and "i < num_holes C"
  shows "subt_at (fill_holes C ts) (hole_poss' C ! i) = ts ! i"
  using assms(1)[symmetric] assms(2)
proof (induct C ts arbitrary: i rule: fill_holes_induct)
  case (MFun f Cs ts)
  have "i < length (concat (map (\<lambda>i. map (op <# i) (hole_poss' (Cs ! i))) [0..<length Cs]))"
    using MFun arg_cong[OF map_nth, of "map num_holes" Cs] by (auto simp: length_concat comp_def)
  thus ?case
    by (auto simp: nth_map[symmetric] map_concat intro!: arg_cong[of _ _ "\<lambda>x. concat x ! _"])
       (insert MFun, auto intro!: nth_equalityI)
qed auto

lemma subm_at_fill_holes_mctxt:
  assumes "length Ds = num_holes C" and "i < num_holes C"
  shows "subm_at (fill_holes_mctxt C Ds) (hole_poss' C ! i) = Ds ! i"
  using assms(1)[symmetric] assms(2)
proof (induct C Ds arbitrary: i rule: fill_holes_induct)
  case (MFun f Cs Ds)
  have "i < length (concat (map (\<lambda>i. map (op <# i) (hole_poss' (Cs ! i))) [0..<length Cs]))"
    using MFun arg_cong[OF map_nth, of "map num_holes" Cs] by (auto simp: length_concat comp_def)
  thus ?case
    by (auto simp: nth_map[symmetric] map_concat intro!: arg_cong[of _ _ "\<lambda>x. concat x ! _"])
       (insert MFun, auto intro!: nth_equalityI)
qed auto

lemma ctxt_of_pos_term_fill_holes:
  assumes "num_holes C = length ts" "i < num_holes C"
  shows "ctxt_of_pos_term (hole_poss' C ! i) (fill_holes C (ts[i := t])) =
    ctxt_of_pos_term (hole_poss' C ! i) (fill_holes C ts)"
  using assms
proof (induct C ts arbitrary: i rule: fill_holes_induct)
  case (MFun f Cs ts)
  then show ?case using partition_by_idx1_bound[of i "map num_holes Cs"] partition_by_idx2_bound[of i "map num_holes Cs"]
  unfolding fill_holes.simps hole_poss'.simps num_holes.simps
    apply (subst (1 2) nth_concat_by_shape, (simp add: comp_def map_upt_len_conv; fail)+)
    apply (subst list_update_concat_by_shape, (simp add: comp_def map_upt_len_conv; fail)+)
    apply (subst (1 2) nth_map, simp)
    apply (subst (1 2) nth_map, simp)
    apply (subst partition_by_concat_id)
    apply (auto intro!: nth_equalityI simp: nth_list_update)
    done
qed auto

lemma hole_poss_in_poss_fill_holes:
  assumes "num_holes C = length ts" "i < num_holes C"
  shows "hole_poss' C ! i \<in> poss (fill_holes C ts)"
proof -
  have "hole_poss C \<subseteq> poss (fill_holes C ts)"
    using all_poss_mctxt_mono[OF fill_holes_suffix[of C ts]] assms by (simp add: all_poss_mctxt_conv)
  then show ?thesis using assms(2) set_hole_poss'[of C] by auto
qed

lemma replace_at_fill_holes:
  assumes "num_holes C = length ts" "i < num_holes C"
  shows "replace_at (fill_holes C ts) (hole_poss' C ! i) ti = fill_holes C (ts[i := ti])"
proof -
  show ?thesis using assms ctxt_supt_id[OF hole_poss_in_poss_fill_holes, of C "ts[i := ti]" i]
    by (simp add: ctxt_of_pos_term_fill_holes subt_at_fill_holes)
qed

lemma fill_holes_rstep_r_p_s':
  assumes "num_holes C = length ss" "i < num_holes C" "(ss ! i, ti) \<in> rstep_r_p_s' \<R> (l, r) p \<sigma>"
  shows "(fill_holes C ss, fill_holes C (ss[i := ti])) \<in> rstep_r_p_s' \<R> (l, r) (hole_poss' C ! i <#> p) \<sigma>"
  using rstep_r_p_s'_mono[OF assms(3), of "ctxt_of_pos_term (hole_poss' C ! i) (fill_holes C ss)"] assms(1,2)
    hole_poss_in_poss_fill_holes[OF assms(1,2)]
  by (simp add: replace_at_fill_holes)

lemma poss_mctxt_append_poss_mctxt:
  "(p <#> q) \<in> poss_mctxt C \<longleftrightarrow> p \<in> all_poss_mctxt C \<and> q \<in> poss_mctxt (subm_at C p)"
  by (induct p arbitrary: C; case_tac C) auto

lemma hole_poss_fill_holes_mctxt:
  assumes "num_holes C = length Ds"
  shows "hole_poss (fill_holes_mctxt C Ds) = {hole_poss' C ! i <#> q |i q. i < length Ds \<and> q \<in> hole_poss (Ds ! i)}"
    (is "?L = ?R")
  using assms
proof -
  have "p \<in> ?L \<Longrightarrow> p \<in> ?R" for p using assms
  proof (induct C Ds arbitrary: p rule: fill_holes_induct)
    case (MFun f Cs Ds)
    obtain i p' where
      *: "i < length Cs" "p = i <# p'"
      "p' \<in> hole_poss (fill_holes_mctxt (Cs ! i) (partition_holes Ds Cs ! i))"
      using MFun(1,3) by (auto)
    with MFun(2)[of i p'] obtain j q where
      "j <  length (partition_holes Ds Cs ! i)" "q \<in> hole_poss (partition_holes Ds Cs ! i ! j)"
      "p' = hole_poss' (Cs ! i) ! j <#> q" by auto
    then show ?case
      using MFun(1) * partition_by_nth_nth(1)[of "map num_holes Cs" "hole_poss' (MFun f Cs)" i j,
          simplified, unfolded length_concat]
        partition_by_concat_id[of "map (\<lambda>i. map (op <# i) (hole_poss' (Cs ! i))) [0..<length Cs]"
          "map num_holes Cs", simplified]
      by (auto intro!: exI[of _ "partition_by_idx (sum_list (map num_holes Cs)) (map num_holes Cs) i j"] exI[of _ q]
        simp: partition_by_nth_nth same_append_eq'[of "PCons _ _", simplified] comp_def map_upt_len_conv)
  qed auto
  moreover have "p \<in> ?R \<Longrightarrow> p \<in> ?L" for p using assms
  proof (induct C Ds arbitrary: p rule: fill_holes_induct)
    case (MFun f Cs Ds)
    obtain i q where *: "i < length Ds" "q \<in> hole_poss (Ds ! i)"
      "p = concat (map (\<lambda>i. map (op <# i) (hole_poss' (Cs ! i))) [0..<length Cs]) ! i <#> q"
      using MFun(1,3) by auto
    let ?i = "map num_holes Cs @\<^sub>1 i" and ?j = "map num_holes Cs @\<^sub>2 i"
    have
      "?i < length Cs" "?j < num_holes (Cs ! ?i)" "q \<in> hole_poss (partition_holes Ds Cs ! ?i ! ?j)"
      "p = ?i <# hole_poss' (Cs ! ?i) ! ?j <#> q"
      using partition_by_idx1_bound[of _ "map num_holes Cs"] * MFun(1)
        partition_by_idx2_bound[of _ "map num_holes Cs"] nth_by_partition_by[of Ds "map num_holes Cs"]
        nth_by_partition_by[of "hole_poss' (MFun f Cs)" "map num_holes Cs" i]
        partition_by_concat_id[of "map (\<lambda>i. map (op <# i) (hole_poss' (Cs ! i))) [0..<length Cs]"
          "map num_holes Cs"]
      by (auto simp: length_concat comp_def map_upt_len_conv )
    then show ?case using MFun(1)
      by (auto intro!: MFun(2)[of ?i "hole_poss' (Cs ! ?i) ! ?j <#> q"] exI[of _ ?j] exI[of _ "q"])
  qed auto
  ultimately show ?thesis by blast
qed

lemma pos_diff_hole_possI:
  "q \<in> hole_poss C \<Longrightarrow> p \<le> q \<Longrightarrow> pos_diff q p \<in> hole_poss (subm_at C p)"
  by (induct C p arbitrary: q rule: subm_at.induct) auto

subsection \<open>An inverse of @{term fill_holes}\<close>

fun unfill_holes :: "('f, 'v) mctxt \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term list" where
  "unfill_holes MHole t = [t]"
| "unfill_holes (MVar w) (Var v) = (if v = w then [] else undefined)"
| "unfill_holes (MFun g Cs) (Fun f ts) = (if f = g \<and> length ts = length Cs then
    concat (map (\<lambda>i. unfill_holes (Cs ! i) (ts ! i)) [0..<length ts]) else undefined)"

lemma length_unfill_holes[simp]:
  assumes "C \<le> mctxt_of_term t"
  shows "length (unfill_holes C t) = num_holes C"
  using assms
proof (induct C t rule: unfill_holes.induct)
  case (3 f Cs g ts) with 3(1)[OF _ nth_mem] 3(2) show ?case
    by (auto simp: less_eq_mctxt_def length_concat
      intro!: cong[of sum_list, OF refl] nth_equalityI elim!: nth_equalityE)
qed (auto simp: less_eq_mctxt_def)

lemma fill_unfill_holes:
  assumes "C \<le> mctxt_of_term t"
  shows "fill_holes C (unfill_holes C t) = t"
  using assms
proof (induct C t rule: unfill_holes.induct)
  case (3 f Cs g ts) with 3(1)[OF _ nth_mem] 3(2) show ?case
    by (auto simp: less_eq_mctxt_def intro!: fill_holes_arbitrary elim!: nth_equalityE)
qed (auto simp: less_eq_mctxt_def split: if_splits)

lemma unfill_fill_holes:
  assumes "length ts = num_holes C"
  shows "unfill_holes C (fill_holes C ts) = ts"
  using assms[symmetric]
proof (induct C ts rule: fill_holes_induct)
  case (MFun f Cs ts) thus ?case
    by (auto intro!: arg_cong[of _ _ concat] nth_equalityI[of _ "partition_holes ts Cs"]
      simp del: concat_partition_by) auto
qed auto

lemma unfill_holes_subt:
  assumes "C \<le> mctxt_of_term t" and "t' \<in> set (unfill_holes C t)"
  shows "t' \<unlhd> t"
  using assms
proof (induct C t rule: unfill_holes.induct)
  case (3 f Cs g ts)
  obtain i where "i < length Cs" and "t' \<in> set (unfill_holes (Cs ! i) (ts ! i))"
    using 3 by (auto dest!: in_set_idx split: if_splits simp: less_eq_mctxt_def)
  then show ?case
    using 3(1)[OF _ nth_mem[of i]] 3(2,3) supteq.subt[of "ts ! i" ts t' g]
    by (auto simp: less_eq_mctxt_def elim!: nth_equalityE split: if_splits)
qed (auto simp: less_eq_mctxt_def split: if_splits)

lemma factor_hole_pos_by_prefix:
  assumes "C \<le> D" "p \<in> hole_poss D"
  obtains q where "q \<le> p" "q \<in> hole_poss C"
  using assms
  by (induct C D arbitrary: p thesis rule: less_eq_mctxt_induct)
    (auto, metis less_eq_simps(4))

lemma unfill_holes_conv:
  assumes "C \<le> mctxt_of_term t"
  shows "unfill_holes C t = map (subt_at t) (hole_poss' C)"
  using assms
proof (induct C t rule: unfill_holes.induct)
  case (3 f Cs g ts) show ?case using 3(2)
    by (auto elim!: less_eq_mctxtE2 simp: map_concat comp_def 3(1) intro!: arg_cong[of _ _ concat])
qed (auto elim: less_eq_mctxtE2)

lemma unfill_holes_by_prefix':
  assumes "num_holes C = length Ds" "fill_holes_mctxt C Ds \<le> mctxt_of_term t"
  shows "unfill_holes (fill_holes_mctxt C Ds) t = concat (map (\<lambda>(D, t). unfill_holes D t) (zip Ds (unfill_holes C t)))"
  using assms
proof (induct C Ds arbitrary: t rule: fill_holes_induct)
  case (MVar v) thus ?case by (cases t) (auto elim: less_eq_mctxtE1)
next
  case (MFun f Cs Ds)
  have [simp]: "length ts = length Cs \<Longrightarrow> map (\<lambda>i. unfill_holes (map (\<lambda>i. fill_holes_mctxt (Cs ! i) (partition_holes Ds Cs ! i))
    [0..<length Cs] ! i) (ts ! i)) [0..<length Cs]
    =  map (\<lambda>i. unfill_holes (fill_holes_mctxt (Cs ! i) (partition_holes Ds Cs ! i)) (ts ! i)) [0..<length Cs]" for ts
    by (auto intro: nth_equalityI)
  obtain ts where lts: "length ts = length Cs" "t = Fun f ts" and
    pre: "i < length Cs \<Longrightarrow> fill_holes_mctxt (Cs ! i) (partition_holes Ds Cs ! i) \<le> mctxt_of_term (ts ! i)" for i
    using MFun(1,3) by (cases t) (auto elim!: less_eq_mctxtE2) 
  have *: "i \<in> set [0..<n] \<Longrightarrow> i < n" for i n by auto
  have **: "(\<And>i. i < n \<Longrightarrow> length (f i) = length (g i)) \<Longrightarrow>
    concat (map (\<lambda>i. zip (f i) (g i)) [0..<n]) = zip (concat (map f [0..<n])) (concat (map g [0..<n]))" for f g n
    by (induct n arbitrary: f g) (auto simp: map_upt_Suc simp del: upt.simps) 
  have ***: "i < length Cs \<Longrightarrow> Cs ! i \<le> mctxt_of_term (ts ! i)" for i
    using fill_holes_mctxt_suffix[of "partition_holes Ds Cs ! i" "Cs ! i", OF length_partition_holes_nth] MFun(1) pre[of i]
    by (auto simp del: fill_holes_mctxt_suffix)
  have [simp]: "concat (map (\<lambda>i. concat (map f (zip (partition_holes Ds Cs ! i) (unfill_holes (Cs ! i) (ts ! i)))))
    [0..<length Cs]) = concat (map f (zip Ds (concat (map (\<lambda>i. unfill_holes (Cs ! i) (ts ! i)) [0..<length Cs]))))" for f
    unfolding concat_map_concat[of "map _ _", unfolded map_map comp_def]
    unfolding map_map[of "map f" "\<lambda>i. zip (_ i) (_ i)", symmetric, unfolded comp_def] map_concat[symmetric]
    using MFun(1) map_nth[of "partition_holes Ds Cs"] by (auto simp: length_unfill_holes[OF ***] **)
  from lts pre show ?case using MFun(1) map_cong[OF refl MFun(2)[OF *], of "[0..<length Cs]" id "\<lambda>i. ts ! i"]    
    by (auto simp del: map_eq_conv)
qed auto

lemma unfill_holes_by_prefix:
  assumes "C \<le> D" and "D \<le> mctxt_of_term t"
  shows "unfill_holes D t = concat (map (\<lambda>p. unfill_holes (subm_at D p) (subt_at t p)) (hole_poss' C))"
  using assms
proof (induct C arbitrary: D t)
  case (MVar x) thus ?case by (cases t) (auto elim!: less_eq_mctxtE1)
next
  case (MFun f Cs)
  obtain Ds where D[simp]: "D = MFun f Ds" "length Ds = length Cs" using MFun(2) by (auto elim: less_eq_mctxtE1)
  obtain ts where t[simp]: "t = Fun f ts" "length ts = length Cs" using MFun(3) by (cases t) (auto elim: less_eq_mctxtE1)
  have "i < length Cs \<Longrightarrow> unfill_holes (Ds ! i) (ts ! i) =
    concat (map (\<lambda>p. unfill_holes (subm_at (Ds ! i) p) (subt_at (ts ! i) p)) (hole_poss' (Cs ! i)))" for i
  proof (intro MFun(1))
    assume "i < length Cs" thus "Cs ! i \<le> Ds ! i" using MFun(2) by (auto elim: less_eq_mctxtE1)
  next
    assume "i < length Cs" thus "Ds ! i \<le> mctxt_of_term (ts ! i)" using MFun(3)
    by (auto elim: less_eq_mctxt_MFunE1)
  qed auto
  then have *: "map (\<lambda>i. unfill_holes (Ds ! i) (ts ! i)) [0..<length Cs] =
        map (concat \<circ> map (\<lambda>p. unfill_holes (subm_at (MFun f Ds) p) (Fun f ts |_ p)) \<circ>
             (\<lambda>i. map (op <# i) (hole_poss' (Cs ! i)))) [0..<length Cs]"
    by (intro nth_equalityI) (auto simp: o_def)
  then show ?case by (auto simp add: map_concat map_map[symmetric] simp del: map_map)
qed auto

lemma unfill_holes_var_subst:
  "C \<le> mctxt_of_term t \<Longrightarrow> unfill_holes (map_vars_mctxt f C) (t \<cdot> (Var \<circ> f)) = map (\<lambda>t. t \<cdot> (Var \<circ> f)) (unfill_holes C t)"
  by (induct C t rule: unfill_holes.induct; (simp only: mctxt_of_term.simps; elim less_eq_mctxtE2)?)
    (auto simp: map_concat intro!: arg_cong[of _ _ concat])

subsection \<open>Ditto for @{term fill_holes_mctxt}\<close>

fun unfill_holes_mctxt :: "('f, 'v) mctxt \<Rightarrow> ('f, 'v) mctxt \<Rightarrow> ('f, 'v) mctxt list" where
  "unfill_holes_mctxt MHole D = [D]"
| "unfill_holes_mctxt (MVar w) (MVar v) = (if v = w then [] else undefined)"
| "unfill_holes_mctxt (MFun g Cs) (MFun f Ds) = (if f = g \<and> length Ds = length Cs then
    concat (map (\<lambda>i. unfill_holes_mctxt (Cs ! i) (Ds ! i)) [0..<length Ds]) else undefined)"

lemma length_unfill_holes_mctxt:
  assumes "C \<le> D"
  shows "length (unfill_holes_mctxt C D) = num_holes C"
  using assms
proof (induct C D rule: unfill_holes_mctxt.induct)
  case (3 f Cs g Ds) with 3(1)[OF _ nth_mem] 3(2) show ?case
    by (auto simp: less_eq_mctxt_def length_concat intro!: cong[of sum_list, OF refl] nth_equalityI elim!: nth_equalityE)
qed (auto simp: less_eq_mctxt_def)

lemma fill_unfill_holes_mctxt:
  assumes "C \<le> D"
  shows "fill_holes_mctxt C (unfill_holes_mctxt C D) = D"
  using assms
proof (induct C D rule: unfill_holes_mctxt.induct)
  case (3 f Cs g Ds) with 3(1)[OF _ nth_mem] 3(2) show ?case
    by (auto simp: less_eq_mctxt_def length_unfill_holes_mctxt intro!: fill_holes_arbitrary elim!: nth_equalityE)
qed (auto simp: less_eq_mctxt_def split: if_splits)

lemma unfill_fill_holes_mctxt:
  assumes "length Ds = num_holes C"
  shows "unfill_holes_mctxt C (fill_holes_mctxt C Ds) = Ds"
  using assms[symmetric]
proof (induct C Ds rule: fill_holes_induct)
  case (MFun f Cs ts) thus ?case
    by (auto intro!: arg_cong[of _ _ concat] nth_equalityI[of _ "partition_holes ts Cs"]
      simp del: concat_partition_by) auto
qed auto

lemma unfill_holes_mctxt_conv:
  assumes "C \<le> D"
  shows "unfill_holes_mctxt C D = map (subm_at D) (hole_poss' C)"
  using assms
proof (induct C D rule: unfill_holes_mctxt.induct)
  case (3 f Cs g Ds) show ?case using 3(2)
    by (auto elim!: less_eq_mctxtE2 simp: map_concat comp_def 3(1) intro!: arg_cong[of _ _ concat])
qed (auto elim: less_eq_mctxtE2)

lemma unfill_holes_mctxt_mctxt_of_term:
  assumes "C \<le> mctxt_of_term t"
  shows "unfill_holes_mctxt C (mctxt_of_term t) = map mctxt_of_term (unfill_holes C t)"
  using assms
proof (induct C arbitrary: t)
  case (MVar x) then show ?case by (cases t) (auto elim: less_eq_mctxtE1)
next
  case MHole then show ?case by (cases t) (auto elim: less_eq_mctxtE1)
next
  case (MFun x1a x2) then show ?case
    by (cases t) (auto elim: less_eq_mctxtE1 simp: map_concat intro!: arg_cong[of _ _ concat])
qed

lemma map_vars_Some_le_mctxt_of_term_mctxt_term_conv:
  "map_vars_mctxt Some C \<le> mctxt_of_term (mctxt_term_conv C)"
  by (induct C) (auto intro: less_eq_mctxtI1)

lemma unfill_holes_map_vars_mctxt_Some_mctxt_term_conv_conv:
 "C \<le> E \<Longrightarrow> unfill_holes (map_vars_mctxt Some C) (mctxt_term_conv E) = map mctxt_term_conv (unfill_holes_mctxt C E)"
  by (induct C E rule: less_eq_mctxt_induct) (auto simp: map_concat intro!: arg_cong[of _ _ concat])

lemma unfill_holes_mctxt_by_prefix':
  assumes "num_holes C = length Ds" "fill_holes_mctxt C Ds \<le> E"
  shows "unfill_holes_mctxt (fill_holes_mctxt C Ds) E = concat (map (\<lambda>(D, E). unfill_holes_mctxt D E) (zip Ds (unfill_holes_mctxt C E)))"
proof -
  have "C \<le> E" using assms(2) fill_holes_mctxt_suffix[OF assms(1)[symmetric]] by auto
  have [simp]: "i < length Ds \<Longrightarrow> Ds ! i \<le> unfill_holes_mctxt C E ! i" for i using assms
    by (subst (asm) fill_unfill_holes_mctxt[OF `C \<le> E`, symmetric])
      (auto simp: less_eq_fill_holes_iff length_unfill_holes_mctxt[OF `C \<le> E`])
  show ?thesis using `C \<le> E` assms
    arg_cong[OF unfill_holes_by_prefix'[of "map_vars_mctxt Some C" "map (map_vars_mctxt Some) Ds" "mctxt_term_conv E"], of "map term_mctxt_conv"]
    order.trans[OF map_vars_mctxt_mono[OF assms(2), of Some] map_vars_Some_le_mctxt_of_term_mctxt_term_conv[of E]]
    unfolding map_vars_mctxt_fill_holes_mctxt[symmetric, OF assms(1)]
    by (auto simp: zip_map1 zip_map2 comp_def unfill_holes_map_vars_mctxt_Some_mctxt_term_conv_conv
      prod.case_distrib map_concat in_set_conv_nth intro!: arg_cong[of _ _ concat])
qed

lemma hole_poss_fill_holes_mctxt_conv:
  assumes "i < num_holes C" "length Cs = num_holes C"
  shows "hole_poss' C ! i \<in> hole_poss (fill_holes_mctxt C Cs) \<longleftrightarrow> Cs ! i = MHole" (is "?L \<longleftrightarrow> ?R")
proof
  assume ?L then show ?R
    using assms arg_cong[OF unfill_holes_mctxt_conv[of C "fill_holes_mctxt C Cs"], of "\<lambda>Cs. Cs ! i"]
    by (auto simp: unfill_fill_holes_mctxt)
next
  assume ?R then show ?L using assms by (force simp: hole_poss_fill_holes_mctxt)
qed

subsection \<open>Function symbols of prefixes\<close>

lemma funas_prefix[simp]:
  "C \<le> D \<Longrightarrow> fn \<in> funas_mctxt C \<Longrightarrow> fn \<in> funas_mctxt D"
  unfolding less_eq_mctxt_def
proof (induct C D rule: inf_mctxt.induct)
  case (4 f Cs g Ds)
  from 4(3) obtain i where "i < length Cs \<and> fn \<in> funas_mctxt (Cs ! i) \<or> fn = (f, length Cs)"
    by (auto dest!: in_set_idx)
  moreover {
    assume "i < length Cs \<and> fn \<in> funas_mctxt (Cs ! i)"
    hence "i < length Ds \<and> fn \<in> funas_mctxt (Ds ! i)" using 4(2)
      by (auto intro!: 4(1)[of _ "Cs ! i" "Ds ! i"] split: if_splits elim!: nth_equalityE simp: in_set_conv_nth)
    hence ?case by (auto)
  }
  ultimately show ?case using 4(2) by auto
qed (auto)

end (* LS_Extras *)
